import argparse
import os
import subprocess
import re
import csv
from pathlib import Path
import sys

parser = argparse.ArgumentParser(description='Run a full benchmark using the CLI and report the results.')
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--old-cli', action='store', help='Path to the CLI of the old DuckDB version to test')
group.add_argument('--versions', type=str, action='store', help='DuckDB versions to test')
parser.add_argument('--new-unittest', action='store', help='Path to the new unittester to run', required=True)
parser.add_argument('--new-cli', action='store', help='Path to the new unittester to run', default=None)
parser.add_argument('--compatibility', action='store', help='Storage compatibility version', default='v1.0.0')
parser.add_argument(
    '--test-config', action='store', help='Test config script to run', default='test/configs/storage_compatibility.json'
)
parser.add_argument('--db-name', action='store', help='Database name to write to', default='bwc_storage_test.db')
parser.add_argument('--abort-on-failure', action='store_true', help='Abort on first failure', default=False)
parser.add_argument('--start-offset', type=int, action='store', help='Test start offset', default=None)
parser.add_argument('--end-offset', type=int, action='store', help='Test end offset', default=None)
parser.add_argument('--no-summarize-failures', action='store_true', help='Skip failure summary', default=False)
parser.add_argument('--list-versions', action='store_true', help='Only list versions to test', default=False)
parser.add_argument(
    '--run-empty-tests',
    action='store_true',
    help='Run tests that don' 't have a CREATE TABLE or CREATE VIEW statement',
    default=False,
)

args, extra_args = parser.parse_known_args()

programs_to_test = []
if args.versions is not None:
    version_splits = args.versions.split('|')
    for version in version_splits:
        cli_path = os.path.join(Path.home(), '.duckdb', 'cli', version, 'duckdb')
        if not os.path.isfile(cli_path):
            if os.system(f'curl https://install.duckdb.org | DUCKDB_VERSION={version} sh'):
                raise Exception(f"CURL install for DuckDB version: {version} failed")
        programs_to_test.append(cli_path)
else:
    programs_to_test.append(args.old_cli)

unittest_program = args.new_unittest
db_name = args.db_name
new_cli = args.new_unittest.replace('test/unittest', 'duckdb') if args.new_cli is None else args.new_cli
summarize_failures = not args.no_summarize_failures

# Use the '-l' parameter to output the list of tests to run
proc = subprocess.run(
    [unittest_program, '--test-config', args.test_config, '-l'] + extra_args,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
)
stdout = proc.stdout.decode('utf8').strip()
stderr = proc.stderr.decode('utf8').strip()
if len(stderr) > 0:
    print("Failed to run program " + unittest_program)
    print("Returncode:", proc.returncode)
    print(stdout)
    print(stderr)
    exit(1)


# The output is in the format of 'PATH\tGROUP', we're only interested in the PATH portion
test_cases = []
first_line = True
for line in stdout.splitlines():
    if first_line:
        first_line = False
        continue
    if len(line.strip()) == 0:
        continue
    splits = line.rsplit('\t', 1)
    test_cases.append(splits[0])

test_cases.sort()
if args.compatibility != 'v1.0.0':
    raise Exception("Only v1.0.0 is supported for now (FIXME)")


def escape_cmd_arg(arg):
    if '"' in arg or '\'' in arg or ' ' in arg or '\\' in arg:
        arg = arg.replace('\\', '\\\\')
        arg = arg.replace('"', '\\"')
        arg = arg.replace("'", "\\'")
        return f'"{arg}"'
    return arg


error_container = []


def handle_failure(test, cmd, msg, stdout, stderr, returncode):
    print(f"==============FAILURE============", file=sys.stderr)
    print(test, file=sys.stderr)
    print(f"==============MESSAGE============", file=sys.stderr)
    print(msg, file=sys.stderr)
    print(f"==============COMMAND============", file=sys.stderr)
    cmd_str = ''
    for entry in cmd:
        cmd_str += escape_cmd_arg(entry) + ' '
    print(cmd_str.strip(), file=sys.stderr)
    print(f"==============RETURNCODE=========", file=sys.stderr)
    print(str(returncode), file=sys.stderr)
    print(f"==============STDOUT=============", file=sys.stderr)
    print(stdout, file=sys.stderr)
    print(f"==============STDERR=============", file=sys.stderr)
    print(stderr, file=sys.stderr)
    print(f"=================================", file=sys.stderr)
    if args.abort_on_failure:
        exit(1)
    else:
        error_container.append({'test': test, 'stderr': stderr})


def run_program(cmd, description):
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = proc.stdout.decode('utf8').strip()
    stderr = proc.stderr.decode('utf8').strip()
    if proc.returncode != 0:
        return {
            'test': test,
            'cmd': cmd,
            'msg': f'Failed to {description}',
            'stdout': stdout,
            'stderr': stderr,
            'returncode': proc.returncode,
        }
    return None


def try_run_program(cmd, description):
    result = run_program(cmd, description)
    if result is None:
        return True
    handle_failure(**result)
    return False


index = 0
start = 0 if args.start_offset is None else args.start_offset
end = len(test_cases) if args.end_offset is None else args.end_offset
for i in range(start, end):
    test = test_cases[i]
    skipped = ''
    if not args.run_empty_tests:
        with open(test, 'r') as f:
            test_contents = f.read().lower()
        if 'create table' not in test_contents and 'create view' not in test_contents:
            skipped = ' (SKIPPED)'

    print(f'[{i}/{len(test_cases)}]: {test}{skipped}')
    if skipped != '':
        continue
    # remove the old db
    try:
        os.remove(db_name)
    except:
        pass
    cmd = [unittest_program, '--test-config', args.test_config, test]
    if not try_run_program(cmd, 'Run Test'):
        print("Failed to Run Test")
        continue

    if not os.path.isfile(db_name):
        # db not created
        print(f"Failed to create a database file by the name of {db_name}")
        continue

    cmd = [
        programs_to_test[-1],
        db_name,
        '-c',
        '.headers off',
        '-csv',
        '-c',
        '.output table_list.csv',
        '-c',
        'SHOW ALL TABLES',
    ]
    if not try_run_program(cmd, 'List Tables'):
        print("Failed to List Tables")
        continue

    tables = []
    with open('table_list.csv', newline='') as f:
        reader = csv.reader(f)
        for row in reader:
            tables.append((row[1], row[2]))
    # no tables / views
    if len(tables) == 0:
        print("No tables/views were created, skipping")
        continue

    # read all tables / views
    failures = []
    for cli in programs_to_test:
        cmd = [cli, db_name]
        for table in tables:
            schema_name = table[0].replace('"', '""')
            table_name = table[1].replace('"', '""')
            cmd += ['-c', f'FROM "{schema_name}"."{table_name}"']
        failure = run_program(cmd, 'Query Tables')
        if failure is not None:
            failures.append(failure)
    if len(failures) > 0:
        # we failed to query the tables
        # this MIGHT be expected - e.g. we might have views that reference stale state (e.g. files that are deleted)
        # try to run it with the new CLI - if this succeeds we have a problem
        new_cmd = [new_cli] + cmd[1:]
        new_failure = run_program(new_cmd, 'Query Tables (New)')
        if new_failure is None:
            # we succeeded with the new CLI - report the failure
            for failure in failures:
                handle_failure(**failure)
        continue

if len(error_container) == 0:
    exit(0)

if summarize_failures:
    print(
        '''\n\n====================================================
================  FAILURES SUMMARY  ================
====================================================\n
''',
        file=sys.stderr,
    )
    for i, error in enumerate(error_container, start=1):
        print(f"\n{i}:", error["test"], "\n", file=sys.stderr)
        print(error["stderr"], file=sys.stderr)

exit(1)



import psycopg2
import argparse
import os
import platform
import shutil
import sys
import subprocess
import multiprocessing.pool

parser = argparse.ArgumentParser(description='Generate TPC-DS reference results from Postgres.')
parser.add_argument(
    '--sf', dest='sf', action='store', help='The TPC-DS scale factor reference results to generate', default=1
)
parser.add_argument(
    '--query-dir',
    dest='query_dir',
    action='store',
    help='The directory with queries to run',
    default='extension/tpcds/dsdgen/queries',
)
parser.add_argument(
    '--answer-dir',
    dest='answer_dir',
    action='store',
    help='The directory where to store the answers',
    default='extension/tpcds/dsdgen/answers/sf${SF}',
)
parser.add_argument(
    '--duckdb-path',
    dest='duckdb_path',
    action='store',
    help='The path to the DuckDB executable',
    default='build/reldebug/duckdb',
)
parser.add_argument(
    '--skip-load',
    dest='skip_load',
    action='store_const',
    const=True,
    help='Whether or not to skip loading',
    default=False,
)
parser.add_argument(
    '--query-list', dest='query_list', action='store', help='The list of queries to run (default = all)', default=''
)
parser.add_argument('--nthreads', dest='nthreads', action='store', type=int, help='The number of threads', default=0)

args = parser.parse_args()

con = psycopg2.connect(database='postgres')
c = con.cursor()
if not args.skip_load:
    tpcds_dir = f'tpcds_sf{args.sf}'

    q = f"""
    CALL dsdgen(sf={args.sf});
    EXPORT DATABASE '{tpcds_dir}' (DELIMITER '|');
    """
    proc = subprocess.Popen([args.duckdb_path, "-c", q])
    proc.wait()
    if proc.returncode != 0:
        exit(1)

    # drop the previous tables
    tables = [
        'name',
        'web_site',
        'web_sales',
        'web_returns',
        'web_page',
        'warehouse',
        'time_dim',
        'store_sales',
        'store_returns',
        'store',
        'ship_mode',
        'reason',
        'promotion',
        'item',
        'inventory',
        'income_band',
        'household_demographics',
        'date_dim',
        'customer_demographics',
        'customer_address',
        'customer',
        'catalog_sales',
        'catalog_returns',
        'catalog_page',
        'call_center',
    ]
    for table in tables:
        c.execute(f'DROP TABLE IF EXISTS {table};')

    with open(os.path.join(tpcds_dir, 'schema.sql'), 'r') as f:
        schema = f.read()

    c.execute(schema)

    with open(os.path.join(tpcds_dir, 'load.sql'), 'r') as f:
        load = f.read()

    load = load.replace(f'{tpcds_dir}/', f'{os.getcwd()}/{tpcds_dir}/')

    c.execute(load)

    con.commit()

# get a list of all queries
queries = os.listdir(args.query_dir)
queries.sort()

answer_dir = args.answer_dir.replace('${SF}', args.sf)

if len(args.query_list) > 0:
    passing_queries = [x + '.sql' for x in args.query_list.split(',')]
    queries = [x for x in queries if x in passing_queries]
    queries.sort()


def run_query(q):
    print(q)
    with open(os.path.join(args.query_dir, q), 'r') as f:
        sql_query = f.read()
    answer_path = os.path.join(os.getcwd(), answer_dir, q.replace('.sql', '.csv'))
    c.execute(f'DROP TABLE IF EXISTS "query_result{q}"')
    c.execute(f'CREATE TABLE "query_result{q}" AS ' + sql_query)
    c.execute(f"COPY \"query_result{q}\" TO '{answer_path}' (FORMAT CSV, DELIMITER '|', HEADER, NULL 'NULL')")


if args.nthreads == 0:
    for q in queries:
        run_query(q)
else:
    pool = multiprocessing.pool.ThreadPool(processes=args.nthreads)

    pool.map(run_query, queries)



import json, os, sys, glob, mimetypes, urllib.request, re

api_url = 'https://api.github.com/repos/duckdb/duckdb/'

if len(sys.argv) < 2:
    print("Usage: [last_tag] ")
    exit(1)


token = os.getenv("GH_TOKEN", "")
if token == "":
    raise ValueError('need a GitHub token in GH_TOKEN')


# amazingly this is the entire code of the pypy package `linkheader-parser`
def extract(link_header):
    """Extract links and their relations from a Link Header Field."""
    links = [l.strip() for l in link_header.split(',')]
    rels = {}
    pattern = r'<(?P<url>.*)>;\s*rel="(?P<rel>.*)"'
    for link in links:
        group_dict = re.match(pattern, link).groupdict()
        rels[group_dict['rel']] = group_dict['url']
    return rels


def gh_api(suburl, full_url=''):
    if full_url == '':
        url = api_url + suburl
    else:
        url = full_url
    headers = {"Content-Type": "application/json", 'Authorization': 'token ' + token}

    req = urllib.request.Request(url, b'', headers)
    req.get_method = lambda: 'GET'
    next_link = None
    try:
        resp = urllib.request.urlopen(req)
        if not resp.getheader("Link") is None:
            link_data = extract(resp.getheader("Link"))
            if "next" in link_data:
                next_link = link_data["next"]
        raw_resp = resp.read().decode()
    except urllib.error.HTTPError as e:
        raw_resp = e.read().decode()  # gah

    ret_json = json.loads(raw_resp)
    if next_link is not None:
        return ret_json + gh_api('', full_url=next_link)
    return ret_json


# get time of tag
old_release = gh_api('releases/tags/%s' % sys.argv[1])
print(old_release["published_at"])

pulls = gh_api('pulls?base=main&state=closed')
for p in pulls:
    if p["merged_at"] is None:
        continue
    if p["merged_at"] < old_release["published_at"]:
        continue
    print(" - #%s: %s" % (p["number"], p["title"]))



import argparse
import re
import sys
from pathlib import Path

GRAMMAR_DIR = Path("extension/autocomplete/grammar/statements")
TRANSFORMER_DIR = Path("extension/autocomplete/transformer")
FACTORY_REG_FILE = Path("extension/autocomplete/transformer/peg_transformer_factory.cpp")
FACTORY_HPP_FILE = Path("extension/autocomplete/include/transformer/peg_transformer.hpp")

# Matches: RuleName <- ...
GRAMMAR_REGEX = re.compile(r"^(\w+)\s*<-")

# Matches: PEGTransformerFactory::TransformRuleName(
TRANSFORMER_REGEX = re.compile(r"PEGTransformerFactory::Transform(\w+)\s*\(")

# Matches: RegisterEnum<...>("RuleName", ...);
ENUM_RULE_REGEX = re.compile(r'RegisterEnum<[^>]+>\s*\(\s*"(\w+)"\s*,')

# Matches: REGISTER_TRANSFORM(TransformRuleName)
REGISTER_TRANSFORM_REGEX = re.compile(r"REGISTER_TRANSFORM\s*\(\s*Transform(\w+)\s*\)")

EXCLUDED_RULES = {
    "FunctionType",
    "IfExists",
    "Database",
    "AbortOrRollback",
    "CommitOrEnd",
    "StartOrBegin",
    "Transaction",
    "VariableAssign",
    "MacroOrFunction",
    "SettingScope",
    "ColLabel",
    "MacroOrFunction",
    "GroupingOrGroupingId",
    "DefaultValues",
    "RowOrRows",
    "Recursive",
    "StarSymbol",
    "IfNotExists",
    "PlainIdentifier",
    "QuotedIdentifier",
    "CreateTableColumnElement",
    "OrReplace",
    "ReservedIdentifier",
    "CatalogName",
    "SchemaName",
    "ReservedSchemaName",
    "ReservedIdentifier",
    "TableName",
    "ReservedTableName",
    "ColumnName",
    "ReservedColumnName",
    "FunctionName",
    "ReservedFunctionName",
    "TableFunctionName",
    "TypeName",
    "PragmaName",
    "SettingName",
    "CopyOptionName",
    "AtTimeZoneOperator",
    "Generated",
    "ColumnConstraint",
    "AlwaysOrByDefault",
    "Lateral",
    "ConstraintNameClause",
    "ReservedSchemaQualification",
    "UsingSample",
    "TableSample",
    "TypeList",
    "NamedParameterAssignment",
    "WithOrdinality",
    "ByName",
    "CollateOperator",
    "ExportClause",
    "ValueOrValues",
    "PivotKeyword",
    "UnpivotKeyword",
    "Unique",
    "DefArg",
    "NoneLiteral",
    "RowOrStruct",
}


def find_grammar_rules(grammar_path):
    """
    Scans the grammar directory for *.gram files and extracts all rule names.

    Returns a dictionary mapping:
    { "filename.gram": (Path, ["Rule1", "Rule2"]), ... }
    """
    all_rules_by_file = {}

    if not grammar_path.is_dir():
        print(f"Error: Grammar directory not found: {grammar_path}", file=sys.stderr)
        sys.exit(1)

    gram_files = sorted(list(grammar_path.glob("*.gram")))
    if not gram_files:
        print(f"Error: No *.gram files found in {grammar_path}", file=sys.stderr)
        sys.exit(1)
    for file_path in gram_files:
        rules_in_file = []
        try:
            with file_path.open("r", encoding="utf-8") as f:
                for line in f:
                    match = GRAMMAR_REGEX.match(line)
                    if match:
                        rules_in_file.append(match.group(1))
        except Exception as e:
            print(f"Error reading {file_path}: {e}", file=sys.stderr)
            continue

        all_rules_by_file[file_path.name] = (file_path, rules_in_file)

    return all_rules_by_file


def find_transformer_rules(transformer_path):
    """
    Scans the transformer directory for *.cpp files and extracts all
    PEGTransformerFactory::Transform...() function implementations.

    Returns a set of all found rule names:
    { "AlterStatement", "AlterTableStmt", ... }
    """
    transformer_rules = set()

    if not transformer_path.is_dir():
        print(
            f"Error: Transformer directory not found: {transformer_path}",
            file=sys.stderr,
        )
        sys.exit(1)

    cpp_files = sorted(list(transformer_path.glob("*.cpp")))
    if not cpp_files:
        print(f"Error: No *.cpp files found in {transformer_path}", file=sys.stderr)
        sys.exit(1)

    print(f"Scanning {len(cpp_files)} transformer files in {transformer_path}...")

    for file_path in cpp_files:
        try:
            with file_path.open("r", encoding="utf-8") as f:
                content = f.read()
                matches = TRANSFORMER_REGEX.finditer(content)
                for match in matches:
                    transformer_rules.add(match.group(1))
        except Exception as e:
            print(f"Error reading {file_path}: {e}", file=sys.stderr)
            continue

    return transformer_rules


def find_factory_registrations(factory_file_path):
    """
    Scans the factory file for RegisterEnum<...> and REGISTER_TRANSFORM(...)

    Returns two sets:
    (enum_rules, registered_rules)
    """
    enum_rules = set()
    registered_rules = set()

    if not factory_file_path.is_file():
        print(f"Error: Factory file not found: {factory_file_path}", file=sys.stderr)
        return enum_rules, registered_rules

    print(f"Scanning factory file: {factory_file_path}...")

    try:
        with factory_file_path.open("r", encoding="utf-8") as f:
            content = f.read()

            # Find enums
            enum_matches = ENUM_RULE_REGEX.finditer(content)
            for match in enum_matches:
                enum_rules.add(match.group(1))

            # Find transformer registrations
            reg_matches = REGISTER_TRANSFORM_REGEX.finditer(content)
            for match in reg_matches:
                registered_rules.add(match.group(1))

    except Exception as e:
        print(f"Error reading {factory_file_path}: {e}", file=sys.stderr)

    return enum_rules, registered_rules


def generate_declaration_stub(rule_name):
    """Generates the C++ method declaration (for the .hpp file)."""
    return f"""// TODO: Verify this return type is correct
static unique_ptr<SQLStatement> Transform{rule_name}(PEGTransformer &transformer, optional_ptr<ParseResult> parse_result);
"""


def generate_registration_stub(rule_name):
    """Generates the C++ registration line (for peg_transformer_factory.cpp)."""
    return f"REGISTER_TRANSFORM(Transform{rule_name});\n"


def generate_implementation_stub(rule_name):
    """Generates the C++ method implementation (for the transform_...cpp file)."""
    return f"""// TODO: Verify this return type is correct
unique_ptr<SQLStatement> PEGTransformerFactory::Transform{rule_name}(PEGTransformer &transformer,
                                                                   optional_ptr<ParseResult> parse_result) {{
	throw NotImplementedException("Transform{rule_name} has not yet been implemented");
}}
"""


def generate_code_for_missing_rules(generation_queue):
    """
    Iterates the generation queue and prints stub code, grouped by rule.
    """
    if not generation_queue:
        print("\nNo missing rules to generate.")
        return

    print("\n--- Code Generation: Missing Stubs ---")
    print("Copy and paste the code below into the correct files.")

    rules_to_generate = []  # List of (rule_name, cpp_filename)
    for cpp_filename, rules in generation_queue.items():
        for rule in rules:
            rules_to_generate.append((rule, cpp_filename))

    # Sort by rule name
    for rule_name, cpp_filename in sorted(rules_to_generate):
        cpp_path = TRANSFORMER_DIR / cpp_filename

        # Constraint: Do not generate code for non-existent files
        if not cpp_path.is_file():
            print(f"\n// --- SKIPPING: {rule_name} (File not found: {cpp_filename}) ---")
            continue

        print(f"--- Generation for rule: {rule_name} ---")
        print(f"1. Add DECLARATION to: {FACTORY_HPP_FILE}")
        print(generate_declaration_stub(rule_name))

        print(f"2. Add REGISTRATION to: {FACTORY_REG_FILE}\nInside the appropriate Register...() function:")
        print(generate_registration_stub(rule_name))

        print(f"3. Add IMPLEMENTATION to: {cpp_path}")
        print(generate_implementation_stub(rule_name))
        print(f"--- End of {rule_name} ---\n")


def main():
    """
    Main script to find rules, compare them, and print a report.
    """
    parser = argparse.ArgumentParser(description="Check transformer coverage and optionally generate stubs.")
    parser.add_argument(
        "-g",
        "--generate",
        action="store_true",
        help="Generate C++ stubs (declaration, registration, implementation) for missing rules.",
    )
    parser.add_argument("-s", "--skip-found", action="store_true", help="Skip output of [ FOUND ] rules")

    args = parser.parse_args()

    grammar_rules_by_file = find_grammar_rules(Path(GRAMMAR_DIR))
    transformer_impls = find_transformer_rules(Path(TRANSFORMER_DIR))
    enum_rules, registered_rules = find_factory_registrations(Path(FACTORY_REG_FILE))

    if not grammar_rules_by_file:
        print("Error: Could not find grammar rules. Exiting.", file=sys.stderr)
        sys.exit(1)

    print("\n--- Rule Coverage Check ---")

    total_grammar_rules = 0
    total_rules_scanned = 0
    total_found_enum = 0
    total_found_registered = 0
    total_missing_registration = 0
    total_missing_implementation = 0
    all_grammar_rules_flat = set()
    missing_rules_by_file = {}

    generation_queue = {}

    # Iterate through each file and its rules
    for file_name, (file_path, grammar_rules) in sorted(grammar_rules_by_file.items()):
        print(f"\n--- File: {file_name} ---")
        missing_count_this_file = 0

        stem = file_path.stem
        cpp_filename = f"transform_{stem}.cpp"
        missing_rules_for_gen = []

        if not grammar_rules:
            print("(No grammar rules found in this file)")
            continue

        for rule_name in sorted(grammar_rules):
            total_rules_scanned += 1
            if rule_name in EXCLUDED_RULES:
                print(f"{'[ EXCLUDED ]':<14} {rule_name}")
                continue

            all_grammar_rules_flat.add(rule_name)
            total_grammar_rules += 1

            is_enum = rule_name in enum_rules
            is_transformer = rule_name in transformer_impls
            is_registered = rule_name in registered_rules

            if is_enum:
                status_str = "[ ENUM ]"
                total_found_enum += 1
            elif is_transformer:
                if is_registered:
                    status_str = "[ FOUND ]"
                    total_found_registered += 1
                else:
                    status_str = "[ NOT REG'D ]"
                    total_missing_registration += 1
                    missing_count_this_file += 1
            else:
                status_str = "[ MISSING ]"
                total_missing_implementation += 1
                missing_count_this_file += 1
                missing_rules_for_gen.append(rule_name)

            if args.skip_found and ("FOUND" in status_str or "ENUM" in status_str):
                continue

            print(f"{status_str:<14} {rule_name}")

        if missing_count_this_file > 0:
            missing_rules_by_file[file_name] = missing_count_this_file

        if missing_rules_for_gen:
            generation_queue[cpp_filename] = missing_rules_for_gen

    total_covered = total_found_enum + total_found_registered
    total_issues = total_missing_implementation + total_missing_registration
    coverage = (total_covered / total_grammar_rules) * 100 if total_grammar_rules > 0 else 0

    print("\n--- Summary: Rule Coverage ---")
    print(f"{'TOTAL RULES SCANNED':<25} : {total_rules_scanned}")
    print(f"  {'  - Excluded':<23} : {len(EXCLUDED_RULES)}")
    print("---------------------------------------")
    print(f"{'TOTAL ACTIONABLE RULES':<25} : {total_grammar_rules}")
    print(f"{'TOTAL COVERED':<25} : {total_covered} ({coverage:.2f}%)")
    print(f"  {'  - Enum':<23} : {total_found_enum}")
    print(f"  {'  - Registered':<23} : {total_found_registered}")
    print(f"{'TOTAL ISSUES':<25} : {total_issues}")

    if missing_rules_by_file:
        print("\n--- Summary: Issues Per File ---")
        for file_name, count in sorted(missing_rules_by_file.items()):
            print(f"{file_name:<25} : {count} issues")

    print("\n--- Orphan / Mismatch Check ---")
    orphan_transformers = transformer_impls - all_grammar_rules_flat - EXCLUDED_RULES
    if orphan_transformers:
        print("\n[!] Orphan Transformer Functions (No matching grammar rule):")
        for rule in sorted(list(orphan_transformers)):
            print(f"  - Transform{rule}")

    orphan_enums = enum_rules - all_grammar_rules_flat - EXCLUDED_RULES
    if orphan_enums:
        print("\n[!] Orphan Enum Rules (No matching grammar rule):")
        for rule in sorted(list(orphan_enums)):
            print(f'  - RegisterEnum("{rule}")')

    orphan_registrations = registered_rules - all_grammar_rules_flat - EXCLUDED_RULES
    if orphan_registrations:
        print("\n[!] Orphan Registrations (No matching grammar rule):")
        for rule in sorted(list(orphan_registrations)):
            print(f"  - REGISTER_TRANSFORM(Transform{rule})")

    missing_impl = registered_rules - transformer_impls
    if missing_impl:
        print("\n  [!] Registered but NOT Implemented (Will cause C++ error):")
        for rule in sorted(list(missing_impl)):
            print(f"  - REGISTER_TRANSFORM(Transform{rule})")

    unnecessary_reg = registered_rules.intersection(enum_rules)
    if unnecessary_reg:
        print("\n[!] Rule registered as BOTH Enum and Transformer (Ambiguous):")
        for rule in sorted(list(unnecessary_reg)):
            print(f"  - {rule}")

    if args.generate:
        generate_code_for_missing_rules(generation_queue)


if __name__ == "__main__":
    main()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest


def test_profiling_json(shell):
    test = (
        ShellTest(shell)
        .statement("PRAGMA enable_profiling=json;")
        .statement('CREATE TABLE "foo"("hello world" INT);')
        .statement("""SELECT "hello world", '\r\t\n\b\f\\' FROM "foo";""")
    )
    result = test.run()
    result.check_stderr(r'"hello world"')
    # This is incorrectly split but that's impossible to do correctly currently.
    result.check_stderr(r''''\r\t"''')
    result.check_stderr(r""""\b\f\\'""")

# fmt: on



import duckdb
import os
import sys

try:
    import pyarrow
    import pyarrow.parquet

    can_run = True
except:
    can_run = False


def generate_header(f):
    f.write(
        '''# name: test/parquet/test_parquet_reader.test
# description: Test Parquet Reader with files on data/parquet-testing
# group: [parquet]

require parquet

statement ok
PRAGMA enable_verification

'''
    )


def get_files():
    files_path = []
    path = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(path, '..', '..')
    os.chdir(path)
    path = os.path.join('data', 'parquet-testing')
    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith(".parquet"):
                files_path.append(os.path.join(root, file))
    return files_path


def get_duckdb_answer(file_path):
    answer = []
    try:
        answer = duckdb.query("SELECT * FROM parquet_scan('" + file_path + "') limit 50").fetchall()
    except Exception as e:
        print(e)
        answer = 'fail'
    return answer


def get_arrow_answer(file_path):
    answer = []
    try:
        arrow = pyarrow.parquet.read_table(file_path)
        duck_rel = duckdb.from_arrow(arrow).limit(50)
        answer = duck_rel.fetchall()
        return answer
    except:
        return 'fail'


def check_result(duckdb_result, arrow_result):
    if arrow_result == 'fail':
        return 'skip'
    if duckdb_result == 'fail':
        return 'fail'
    if duckdb_result != arrow_result:
        return 'fail'
    return 'pass'


def sanitize_string(s):
    return str(s).replace('None', 'NULL').replace("b'", "").replace("'", "")


def result_to_string(arrow_result):
    result = ''
    for row_idx in range(len(arrow_result)):
        for col_idx in range(len(arrow_result[0])):
            value = arrow_result[row_idx][col_idx]
            if isinstance(value, dict):
                items = [f"'{k}': {sanitize_string(v)}" for k, v in value.items()]  # no quotes
                value = "{" + ", ".join(items) + "}"
                print(type(value), value)
            else:
                value = sanitize_string(value)
            result += value + "\t"
        result += "\n"
    result += "\n"
    return result


def generate_parquet_test_body(result, arrow_result, file_path):
    columns = 'I' * len(arrow_result[0])
    test_body = "query " + columns + "\n"
    test_body += "SELECT * FROM parquet_scan('" + file_path + "') limit 50 \n"
    test_body += "----\n"
    test_body += result_to_string(arrow_result)
    return test_body


def generate_test(file_path):
    duckdb_result = get_duckdb_answer(file_path)
    arrow_result = get_arrow_answer(file_path)
    result = check_result(duckdb_result, arrow_result)
    test_body = ""
    if result == 'skip':
        return
    if result == 'fail':
        test_body += "mode skip \n\n"
        test_body += generate_parquet_test_body(result, arrow_result, file_path)
        test_body += "mode unskip \n\n"
    else:
        test_body += generate_parquet_test_body(result, duckdb_result, file_path)
    return test_body


def generate_body(f):
    files_path = get_files()
    for file in files_path:
        print(file)
        test_body = generate_test(file)
        if test_body != None:
            f.write(test_body)


f = open("test_parquet_reader.test", "w")

generate_header(f)
generate_body(f)


f.close()



#! /usr/bin/env python3

import itertools
import numpy as np
from numpy.random import Generator, PCG64
from pathlib import Path
import pyarrow as pa
import pyarrow.parquet as pq


"""Generate data to test parquet data page decompression."""


COMPRESSION_CODECS = [
    "NONE",
    "SNAPPY",
    "GZIP",
    # Brotli is currently not supported by duckdb
    "BROTLI",
    # This generates the new LZ4_RAW parquet compression, which duckdb does not
    # support
    "LZ4",
    "ZSTD",
]


DATA_PAGE_VERSIONS = [
    "1.0",
    "2.0",
]


def build_table():
    # Init rng in a reproducible way
    rng = Generator(PCG64(12345))


    # Plain table.
    N = 30  # column count
    p = .2  # NULL probability

    columns = {}

    # Integer columns, no nesting, no NULL, no repetition
    columns["plain"] = pa.array(np.arange(N))
    columns["plain_random"] = pa.array(rng.choice(N, N))

    # Mixed dtype struct column, NULLs exist at all levels
    x = pa.array(
        rng.choice(["foo", "bar", "baz"], N),
        mask=rng.choice([True, False], N, p=[p, 1 - p]),
    )
    y = pa.array(
        rng.choice(42, N),
        mask=rng.choice([True, False], N, p=[p, 1 - p]),
    )
    z = pa.StructArray.from_arrays(
        (x, y), names=("string", "int"),
        mask=pa.array(rng.choice([True, False], N, p=[p, 1 - p])),
    )
    columns["nested_nulls"] = z

    # Integer list with variable list length and NULLs
    values = list(range(42)) + [None]
    columns["list"] = pa.array(
        [rng.choice(values, count) for count in rng.choice(20, N)],
        mask=pa.array(rng.choice([True, False], N, p=[p, 1 - p])),
    )

    return pa.Table.from_pydict(columns)


table = build_table()

root = Path("generated")
root.mkdir(exist_ok=True)

for compression, data_page_version in itertools.product(COMPRESSION_CODECS, DATA_PAGE_VERSIONS):
    pq_args = { 
        "data_page_version": data_page_version,
        "compression": compression,
    }

    pq.write_table(
        table,
        (root / "_".join([
            f"data_page={data_page_version[0]}",
            f"{compression}",
        ])).with_suffix(".parquet"),
        **pq_args
    )



# this script re-generates the binary file used for Test deserialized plans from file
# before running this script, increment the version number in src/planner/logical_operator.cpp and
# recompile (make debug)
# Note that the test is not linked unless you BUILD_TPCH=1

import os
import subprocess
from python_helpers import open_utf8

shell_proc = os.path.join('build', 'debug', 'test', 'unittest')
gen_binary_file = os.path.join('test', 'api', 'serialized_plans', 'serialized_plans.binary')


def try_remove_file(fname):
    try:
        os.remove(fname)
    except:
        pass


try_remove_file(gen_binary_file)


def run_test(test):
    print(test)
    env = os.environ.copy()
    env["GEN_PLAN_STORAGE"] = "1"
    res = subprocess.run([shell_proc, test], capture_output=True, env=env)
    stdout = res.stdout.decode('utf8').strip()
    stderr = res.stderr.decode('utf8').strip()
    if res.returncode != 0:
        print("Failed to create binary file!")
        print("----STDOUT----")
        print(stdout)
        print("----STDERR----")
        print(stderr)


run_test("Generate serialized plans file")



import tempfile
import os
from pyspark.sql import SparkSession
from pyspark.sql.types import *
import glob

parquet_compression = 'snappy'

outdir = tempfile.mkdtemp()
parquet_folder = os.path.join(outdir, "out.parquet")
nthreads = 8
memory_gb = 10

spark = SparkSession.builder.master("local[%d]" % nthreads).config('spark.sql.parquet.compression.codec', parquet_compression).config("spark.ui.enabled", "false").config("spark.local.dir", outdir).config("spark.driver.memory", "%dg" % memory_gb).config("spark.executor.memory", "%dg" % memory_gb).getOrCreate()
sc = spark.sparkContext


schema = StructType([
    StructField("l_orderkey",      LongType(),    False),
    StructField("l_partkey",       LongType(),    False),
    StructField("l_suppkey",       LongType(),    False),

    StructField("l_linenumber",    IntegerType(), False),
    StructField("l_quantity",      IntegerType(), False),

    StructField("l_extendedprice", DoubleType(),  False),
    StructField("l_discount",      DoubleType(),  False),
    StructField("l_tax",           DoubleType(),  False),
 
    StructField("l_returnflag",    StringType(),  False),
    StructField("l_linestatus",    StringType(),  False),

    StructField("l_shipdate",      StringType(),  False),
    StructField("l_commitdate",    StringType(),  False),
    StructField("l_receiptdate",   StringType(),  False),

    StructField("l_shipinstruct",  StringType(),  False),
    StructField("l_shipmode",      StringType(),  False),
    StructField("l_comment",       StringType(),  False)])

df = spark.read.format("csv").schema(schema).option("header", "false").option("delimiter", "|").load("lineitem-sf1.tbl.gz").repartition(1)
df.write.mode('overwrite').format("parquet").save(parquet_folder)

os.rename(glob.glob(os.path.join(parquet_folder, '*.parquet'))[0], "lineitem-sf1.%s.parquet" % parquet_compression)



import os
import sys
import time

if len(sys.argv) <= 1:
    print("Expected usage: python3 repeat_until_success.py [command]")
    exit(1)

ntries = 10
sleep_duration = 3
cmd = sys.argv[1]

for i in range(ntries):
    ret = os.system(cmd)
    if ret is None or ret == 0:
        exit(0)
    print("Command {{ " + cmd + " }} failed, retrying (" + str(i + 1) + "/" + str(ntries) + ")")
    time.sleep(sleep_duration)

exit(1)



import argparse
import os
import math
import re

parser = argparse.ArgumentParser(description='Check code coverage results')

parser.add_argument(
    '--uncovered_files',
    action='store',
    help='Set of files that are not 100% covered',
    default=os.path.join(".github", "config", "uncovered_files.csv"),
)
parser.add_argument('--directory', help='Directory of generated HTML files', action='store', default='coverage_html')
parser.add_argument('--fix', help='Fill up the uncovered_files.csv with all files', action='store_true', default=False)


args = parser.parse_args()
if not os.path.exists(args.directory):
    print(f"The provided directory ({args.directory}) does not exist, please create it first")
    exit(1)

covered_regex = (
    r'<a name="(\d+)">[ \t\n]*<span class="lineNum">[ \t\n0-9]+</span><span class="{COVERED_CLASS}">[ \t\n0-9]+:([^<]+)'
)


def get_original_path(path):
    return (
        path.replace('.gcov.html', '')
        .replace(os.getcwd(), '')
        .replace('coverage_html' + os.path.sep, '')
        .replace('home/runner/work/duckdb/duckdb/', '')
    )


def cleanup_line(line):
    return line.replace('&amp;', '&').replace('&lt;', '<').replace('&gt;', '>').replace('&quot;', '"')


partial_coverage_dict = {}
with open(args.uncovered_files, 'r') as f:
    for line in f.readlines():
        splits = line.split('\t')
        partial_coverage_dict[splits[0]] = int(splits[1].strip())

if args.fix:
    uncovered_file = open(args.uncovered_files, 'w+')

DASH_COUNT = 80
total_difference = 0
allowed_difference = 0


def check_file(path, partial_coverage_dict):
    global any_failed
    global total_difference
    if not '.cpp' in path and not '.hpp' in path:
        # files are named [path].[ch]pp
        return
    if not '.html' in path:
        return
    with open(path, 'r') as f:
        text = f.read()
    original_path = get_original_path(path)
    uncovered_lines = re.findall(covered_regex.replace('{COVERED_CLASS}', 'lineNoCov'), text)
    covered_lines = re.findall(covered_regex.replace('{COVERED_CLASS}', 'lineCov'), text)

    total_lines = len(uncovered_lines) + len(covered_lines)
    if total_lines == 0:
        # no lines to cover - skip
        return

    coverage_percentage = round(len(covered_lines) / (total_lines) * 100, 2)
    expected_uncovered_lines = 0
    if original_path in partial_coverage_dict:
        expected_uncovered_lines = partial_coverage_dict[original_path]
    if args.fix:
        if expected_uncovered_lines == 0 and len(uncovered_lines) == 0:
            return
        expected_uncovered = max(expected_uncovered_lines, len(uncovered_lines) + 1)
        uncovered_file.write(f'{original_path}\t{expected_uncovered}\n')
        return

    if len(uncovered_lines) > expected_uncovered_lines:
        total_difference += len(uncovered_lines) - expected_uncovered_lines

        print("-" * DASH_COUNT)
        print(f"Coverage failure in file {original_path}")
        print("-" * DASH_COUNT)
        print(f"Coverage percentage: {coverage_percentage}%")
        print(f"Uncovered lines: {len(uncovered_lines)}")
        print(f"Covered lines: {len(covered_lines)}")
        print("-" * DASH_COUNT)
        print(f"Expected uncovered lines: {expected_uncovered_lines}")
        print("-" * DASH_COUNT)
        print("Uncovered lines")
        print("-" * DASH_COUNT)
        for e in uncovered_lines:
            print(e[0] + ' ' * 8 + cleanup_line(e[1]))


def scan_directory(path):
    file_list = []
    if os.path.isfile(path):
        file_list.append(path)
    else:
        files = os.listdir(path)
        for file in files:
            file_list += scan_directory(os.path.join(path, file))
    return file_list


files = scan_directory(args.directory)
files.sort()

for file in files:
    check_file(file, partial_coverage_dict)

if args.fix:
    uncovered_file.close()

if total_difference > allowed_difference:
    exit(1)
elif total_difference > 0:
    print("-" * DASH_COUNT)
    print("SUCCESS-ish")
    print("-" * DASH_COUNT)
    print(f"{total_difference} lines were uncovered but this falls within the margin of {allowed_difference}")



# Taken from https://github.com/opengeospatial/geoparquet/tree/main/test_data

"""
Generates example data using pyarrow by running `python generate_test_data.py`.

You can print the metadata with:

.. code-block:: python

   >>> import json, pprint, pyarrow.parquet as pq
   >>> pprint.pprint(json.loads(pq.read_schema("example.parquet").metadata[b"geo"]))
"""

import json
import pathlib
import copy

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from pyarrow.csv import write_csv

from shapely import from_wkt, to_wkb


HERE = pathlib.Path(__file__).parent


metadata_template = {
    "version": "1.1.0",
    "primary_column": "geometry",
    "columns": {
        "geometry": {
            "encoding": "WKB",
            "geometry_types": [],
        },
    },
}


## Various geometry types with WKB and native (GeoArrow-based) encodings


def write_encoding_files(geometries_wkt, geometries_geoarrow, geometry_type):

    table = pa.table({"col": range(len(geometries_wkt)), "geometry": geometries_wkt})
    write_csv(table, HERE / f"data-{geometry_type.lower()}-wkt.csv")

    # WKB encoding
    table = pa.table({"col": range(len(geometries_wkt)), "geometry": to_wkb(from_wkt(geometries_wkt))})
    metadata = copy.deepcopy(metadata_template)
    metadata["columns"]["geometry"]["geometry_types"] = [geometry_type]
    table = table.replace_schema_metadata({"geo": json.dumps(metadata)})
    pq.write_table(table, HERE / f"data-{geometry_type.lower()}-encoding_wkb.parquet")

    # native (geoarrow) encoding
    table = pa.table({"col": range(len(geometries_wkt)), "geometry": geometries_geoarrow})
    metadata["columns"]["geometry"]["encoding"] = geometry_type.lower()
    table = table.replace_schema_metadata({"geo": json.dumps(metadata)})
    pq.write_table(table, HERE / f"data-{geometry_type.lower()}-encoding_native.parquet")


# point

geometries_wkt = [
    "POINT (30 10)",
    "POINT EMPTY",
    None,
    "POINT (40 40)",
]

point_type = pa.struct([pa.field("x", pa.float64(), nullable=False), pa.field("y", pa.float64(), nullable=False)])
geometries = pa.array(
    [(30, 10), (float("nan"), float("nan")), (float("nan"), float("nan")), (40, 40)],
    mask=np.array([False, False, True, False]),
    type=point_type,
)

write_encoding_files(geometries_wkt, geometries, geometry_type="Point")

# linestring

geometries_wkt = ["LINESTRING (30 10, 10 30, 40 40)", "LINESTRING EMPTY", None]

linestring_type = pa.list_(pa.field("vertices", point_type, nullable=False))
geometries = pa.array(
    [[(30, 10), (10, 30), (40, 40)], [], []], mask=np.array([False, False, True]), type=linestring_type
)

write_encoding_files(geometries_wkt, geometries, geometry_type="LineString")

# polygon

geometries_wkt = [
    "POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))",
    "POLYGON ((35 10, 45 45, 15 40, 10 20, 35 10), (20 30, 35 35, 30 20, 20 30))",
    "POLYGON EMPTY",
    None,
]

polygon_type = pa.list_(pa.field("rings", pa.list_(pa.field("vertices", point_type, nullable=False)), nullable=False))
geometries = pa.array(
    [
        [[(30, 10), (40, 40), (20, 40), (10, 20), (30, 10)]],
        [[(35, 10), (45, 45), (15, 40), (10, 20), (35, 10)], [(20, 30), (35, 35), (30, 20), (20, 30)]],
        [],
        [],
    ],
    mask=np.array([False, False, False, True]),
    type=polygon_type,
)

write_encoding_files(geometries_wkt, geometries, geometry_type="Polygon")

# multipoint

geometries_wkt = [
    "MULTIPOINT ((30 10))",
    "MULTIPOINT ((10 40), (40 30), (20 20), (30 10))",
    "MULTIPOINT EMPTY",
    None,
]

multipoint_type = pa.list_(pa.field("points", point_type, nullable=False))
geometries = pa.array(
    [
        [(30, 10)],
        [(10, 40), (40, 30), (20, 20), (30, 10)],
        [],
        [],
    ],
    mask=np.array([False, False, False, True]),
    type=multipoint_type,
)

write_encoding_files(geometries_wkt, geometries, geometry_type="MultiPoint")

# multilinestring

geometries_wkt = [
    "MULTILINESTRING ((30 10, 10 30, 40 40))",
    "MULTILINESTRING ((10 10, 20 20, 10 40), (40 40, 30 30, 40 20, 30 10))",
    "MULTILINESTRING EMPTY",
    None,
]

multilinestring_type = pa.list_(pa.field("linestrings", linestring_type, nullable=False))
geometries = pa.array(
    [
        [[(30, 10), (10, 30), (40, 40)]],
        [[(10, 10), (20, 20), (10, 40)], [(40, 40), (30, 30), (40, 20), (30, 10)]],
        [],
        [],
    ],
    mask=np.array([False, False, False, True]),
    type=multilinestring_type,
)

write_encoding_files(geometries_wkt, geometries, geometry_type="MultiLineString")

# multipolygon

geometries_wkt = [
    "MULTIPOLYGON (((30 10, 40 40, 20 40, 10 20, 30 10)))",
    "MULTIPOLYGON (((30 20, 45 40, 10 40, 30 20)), ((15 5, 40 10, 10 20, 5 10, 15 5)))",
    "MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)), ((20 35, 10 30, 10 10, 30 5, 45 20, 20 35), (30 20, 20 15, 20 25, 30 20)))",
    "MULTIPOLYGON EMPTY",
    None,
]

multipolygon_type = pa.list_(pa.field("polygons", polygon_type, nullable=False))
geometries = pa.array(
    [
        [[[(30, 10), (40, 40), (20, 40), (10, 20), (30, 10)]]],
        [[[(30, 20), (45, 40), (10, 40), (30, 20)]], [[(15, 5), (40, 10), (10, 20), (5, 10), (15, 5)]]],
        [
            [[(40, 40), (20, 45), (45, 30), (40, 40)]],
            [[(20, 35), (10, 30), (10, 10), (30, 5), (45, 20), (20, 35)], [(30, 20), (20, 15), (20, 25), (30, 20)]],
        ],
        [],
        [],
    ],
    mask=np.array([False, False, False, False, True]),
    type=multipolygon_type,
)

write_encoding_files(geometries_wkt, geometries, geometry_type="MultiPolygon")



import json
from collections import defaultdict
import sys


def load_skip_dict(path):
    """Load a skip_tests list into a dict: reason -> set(paths)."""
    with open(path) as f:
        data = json.load(f)

    out = {}
    for block in data.get("skip_tests", []):
        reason = block["reason"]
        paths = set(block.get("paths", []))
        out[reason] = paths
    return out


def compare_files(file_a, file_b):
    a = load_skip_dict(file_a)
    b = load_skip_dict(file_b)

    all_reasons = set(a.keys()) | set(b.keys())

    for reason in sorted(all_reasons):
        paths_a = a.get(reason, set())
        paths_b = b.get(reason, set())

        added = sorted(paths_b - paths_a)
        removed = sorted(paths_a - paths_b)

        if not added and not removed:
            continue

        print(f"\n=== Reason: {reason} ===")

        if removed:
            print("  - Present in A but NOT in B:")
            for p in removed:
                print(f"      {p}")

        if added:
            print("  + Present in B but NOT in A:")
            for p in added:
                print(f"      {p}")


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: python compare_skip_tests.py <fileA.json> <fileB.json>")
        exit(1)

    compare_files(sys.argv[1], sys.argv[2])



import json
import os
import sys

# Pass vcpkg.json files to merge their dependencies and produce a single vcpkg.json with their
# combined & deduplicated dependencies. Note that this script is very dumb and some manual merging may be required
# to combine extensions from multiple builds in the case of colliding dependencies.

dependencies_str = []
dependencies_dict = []
merged_overlay_ports = []
merged_overlay_triplets = []


def prefix_overlay_ports_or_triples(overlay_dir, path_to_vcpkg_json):
    def prefix_overlay_port_or_triplet(overlay_port_or_triplet):
        vcpkg_prefix_path = path_to_vcpkg_json[0 : path_to_vcpkg_json.find("/vcpkg.json")]
        if len(vcpkg_prefix_path) == 0:
            return overlay_port_or_triplet
        return vcpkg_prefix_path + '/' + overlay_port_or_triplet

    return map(prefix_overlay_port_or_triplet, overlay_dir)


for file in sys.argv[1:]:
    f = open(file)
    data = json.load(f)

    if 'dependencies' in data:
        for dep in data['dependencies']:
            if type(dep) is str:
                dependencies_str.append(dep)
            elif type(dep) is dict:
                dependencies_dict.append(dep)
            else:
                raise Exception(f"Unknown entry type found in dependencies: '{dep}'")

    if 'vcpkg-configuration' in data:
        if 'overlay-ports' in data['vcpkg-configuration']:
            merged_overlay_ports += prefix_overlay_ports_or_triples(data['vcpkg-configuration']['overlay-ports'], file)
        if 'overlay-triplets' in data['vcpkg-configuration']:
            merged_overlay_triplets += prefix_overlay_ports_or_triples(
                data['vcpkg-configuration']['overlay-triplets'], file
            )

final_deduplicated_deps = list()
dedup_set = set()

for dep in dependencies_dict:
    if dep['name'] not in dedup_set:
        final_deduplicated_deps.append(dep)
        # TODO: deduplication is disabled for now, just let vcpkg handle duplicates in deps
        # dedup_set.add(dep['name'])

for dep in dependencies_str:
    if dep not in dedup_set:
        final_deduplicated_deps.append(dep)
        # TODO: deduplication is disabled for now, just let vcpkg handle duplicates in deps
        # dedup_set.add(dep)

opensslVersion = os.getenv("OPENSSL_VERSION_OVERRIDE", "3.6.0")
data = {
    "description": f"Auto-generated vcpkg.json for combined DuckDB extension build, generated by 'scripts/merge_vcpkg_deps.py'",
    "builtin-baseline": "84bab45d415d22042bd0b9081aea57f362da3f35",
    "dependencies": final_deduplicated_deps,
    "overrides": [{"name": "openssl", "version": opensslVersion}],
}

data['vcpkg-configuration'] = {}

if merged_overlay_ports:
    data['vcpkg-configuration']['overlay-ports'] = merged_overlay_ports

if merged_overlay_triplets:
    data['vcpkg-configuration']['overlay-triplets'] = merged_overlay_triplets

REGISTRY_BASELINE = 'f13de79b66345fe31d3f9c119517613023d194a8'
# NOTE: use 'scripts/list_vcpkg_registry_packages.py --baseline <baseline>' to generate the list of packages
data['vcpkg-configuration']['registries'] = [
    {
        "kind": "git",
        "repository": "https://github.com/duckdb/vcpkg-duckdb-ports",
        "baseline": REGISTRY_BASELINE,
        "packages": ['avro-c', 'vcpkg-cmake'],
    }
]

# Print output
print("Writing to 'build/extension_configuration/vcpkg.json': ")
print(data["dependencies"])

with open('build/extension_configuration/vcpkg.json', 'w', encoding='utf-8') as f:
    json.dump(data, f, ensure_ascii=False, indent=4)



import json
import os

scripts_dir = os.path.dirname(os.path.abspath(__file__))
VERSION_MAP_PATH = scripts_dir + "/../src/storage/version_map.json"
STORAGE_INFO_PATH = scripts_dir + "/../src/storage/storage_info.cpp"
START_MARKER = "// START OF {type} VERSION INFO"
END_MARKER = "// END OF {type} VERSION INFO"


def generate_version_info_array(storage_versions, type, name, default):
    result = []
    name_upper = name.upper()
    if 'latest' in storage_versions:
        latest_value = storage_versions['latest']
        result.append(f"const uint64_t LATEST_{name_upper} = {latest_value};")

    result.append(f"const uint64_t DEFAULT_{name_upper} = {default};")

    result.append(f"static const {type} {name}[] = {{")

    for version_name, storage_version in storage_versions.items():
        result.append(f'\t{{"{version_name}", {storage_version}}},')

    result.append("\t{nullptr, 0}")
    result.append("};\n")

    return "\n".join(result)


def main():

    with open(VERSION_MAP_PATH, 'r') as json_file:
        version_map = json.load(json_file)

    with open(STORAGE_INFO_PATH, "r") as cpp_file:
        content = cpp_file.read()

    for key in version_map['serialization']['values'].keys():
        if key in ['latest']:
            continue
        if key not in version_map['storage']['values'].keys():
            print(f'Key {key} found in serialization version but not in storage version')
            exit(1)
    types = ['storage', 'serialization']
    for type in version_map:
        if type not in types:
            print("Unexpected key {type}")
            exit(1)
        capitalized_type = type.capitalize()
        upper_type = type.upper()
        array_code = generate_version_info_array(
            version_map[type]['values'],
            f'{capitalized_type}VersionInfo',
            f'{type}_version_info',
            version_map[type]['default'],
        )

        start_marker = START_MARKER.format(type=upper_type)
        start_index = content.find(start_marker)
        if start_index == -1:
            print(f"storage_info.cpp is corrupted, could not find the START_MARKER for {type}")
            exit(1)

        end_marker = END_MARKER.format(type=upper_type)
        end_index = content.find(end_marker)
        if end_index == -1:
            print(f"storage_info.cpp is corrupted, could not find the END_MARKER for {type}")
            exit(1)
        content = content[: start_index + len(start_marker)] + "\n" + array_code + content[end_index:]

    with open(STORAGE_INFO_PATH, "w") as cpp_file:
        cpp_file.write(content)


if __name__ == "__main__":
    main()



import sqllogictest
from sqllogictest import SQLParserException, SQLLogicParser, SQLLogicTest
import duckdb
from typing import Optional
import argparse
import shutil
import os
import subprocess

# example usage: python3 scripts/test_serialization_bwc.py --old-source ../duckdb-bugfix --test-file test/sql/aggregate/aggregates/test_median.test

serialized_path = os.path.join('test', 'api', 'serialized_plans')
db_load_path = os.path.join(serialized_path, 'db_load.sql')
queries_path = os.path.join(serialized_path, 'queries.sql')
result_binary = os.path.join(serialized_path, 'serialized_plans.binary')
unittest_binary = os.path.join('build', 'debug', 'test', 'unittest')


def complete_query(q):
    q = q.strip()
    if q.endswith(';'):
        return q
    return q + ';'


def parse_test_file(filename):
    parser = SQLLogicParser()
    try:
        out: Optional[SQLLogicTest] = parser.parse(filename)
        if not out:
            raise SQLParserException(f"Test {filename} could not be parsed")
    except:
        return {'load': [], 'query': []}
    loop_count = 0
    load_statements = []
    query_statements = []
    for stmt in out.statements:
        if type(stmt) is sqllogictest.statement.skip.Skip:
            # mode skip - just skip entire test
            break
        if type(stmt) is sqllogictest.statement.loop.Loop or type(stmt) is sqllogictest.statement.foreach.Foreach:
            loop_count += 1
        if type(stmt) is sqllogictest.statement.endloop.Endloop:
            loop_count -= 1
        if loop_count > 0:
            # loops are ignored currently
            continue
        if not (
            type(stmt) is sqllogictest.statement.query.Query or type(stmt) is sqllogictest.statement.statement.Statement
        ):
            # only handle query and statement nodes for now
            continue
        if type(stmt) is sqllogictest.statement.statement.Statement:
            # skip expected errors
            if stmt.expected_result.type == sqllogictest.ExpectedResult.Type.ERROR:
                continue
        query = ' '.join(stmt.lines)
        try:
            sql_stmt_list = duckdb.extract_statements(query)
        except KeyboardInterrupt:
            raise
        except:
            continue
        for sql_stmt in sql_stmt_list:
            if sql_stmt.type == duckdb.StatementType.SELECT:
                query_statements.append(query)
            elif sql_stmt.type == duckdb.StatementType.PRAGMA:
                continue
            else:
                load_statements.append(query)
    return {'load': load_statements, 'query': query_statements}


def build_sources(old_source, new_source):
    # generate the sources
    current_path = os.getcwd()
    os.chdir(old_source)
    # build if not yet build
    if not os.path.isfile(unittest_binary):
        res = subprocess.run(['make', 'debug']).returncode
        if res != 0:
            raise Exception("Failed to build old sources")

    # run the verification
    os.chdir(current_path)
    os.chdir(new_source)

    # build if not yet build
    if not os.path.isfile(unittest_binary):
        res = subprocess.run(['make', 'debug']).returncode
        if res != 0:
            raise Exception("Failed to build new sources")
    os.chdir(current_path)


def run_test(filename, old_source, new_source, no_exit):
    statements = parse_test_file(filename)

    # generate the sources
    current_path = os.getcwd()
    os.chdir(old_source)
    # write the files
    with open(os.path.join(old_source, db_load_path), 'w+') as f:
        for stmt in statements['load']:
            f.write(complete_query(stmt) + '\n')

    with open(os.path.join(old_source, queries_path), 'w+') as f:
        for stmt in statements['query']:
            f.write(complete_query(stmt) + '\n')

    # generate the serialization
    my_env = os.environ.copy()
    my_env['GEN_PLAN_STORAGE'] = '1'
    res = subprocess.run(['build/debug/test/unittest', 'Generate serialized plans file'], env=my_env).returncode
    if res != 0:
        print(f"SKIPPING TEST {filename}")
        return True

    os.chdir(current_path)

    # copy over the files
    for f in [db_load_path, queries_path, result_binary]:
        shutil.copy(os.path.join(old_source, f), os.path.join(new_source, f))

    # run the verification
    os.chdir(new_source)

    res = subprocess.run(['build/debug/test/unittest', "Test deserialized plans from file"]).returncode
    if res != 0:
        if no_exit:
            print("BROKEN TEST")
            with open('broken_tests.list', 'a') as f:
                f.write(filename + '\n')
            return False
        raise Exception("Deserialization failure")
    os.chdir(current_path)
    return True


def parse_excluded_tests(path):
    exclusion_list = {}
    with open(path) as f:
        for line in f:
            if len(line.strip()) == 0 or line[0] == '#':
                continue
            exclusion_list[line.strip()] = True
    return exclusion_list


def find_tests_recursive(dir, excluded_paths):
    test_list = []
    for f in os.listdir(dir):
        path = os.path.join(dir, f)
        if path in excluded_paths:
            continue
        if os.path.isdir(path):
            test_list += find_tests_recursive(path, excluded_paths)
        elif path.endswith('.test'):
            test_list.append(path)
    return test_list


def main():
    parser = argparse.ArgumentParser(description="Test serialization")
    parser.add_argument("--new-source", type=str, help="Path to the new source", default='.')
    parser.add_argument("--old-source", type=str, help="Path to the old source")
    parser.add_argument("--start-at", type=str, help="Start running tests at this specific test", default=None)
    parser.add_argument("--no-exit", action="store_true", help="Keep running even if a test fails", default=False)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--test-file", type=str, help="Path to the SQL logic file", default='')
    group.add_argument("--all-tests", action='store_true', help="Run all tests", default=False)
    group.add_argument("--test-list", type=str, help="Load tests to run from a file list", default=None)
    args = parser.parse_args()

    old_source = args.old_source
    new_source = args.new_source
    files = []
    if args.all_tests:
        # run all tests
        excluded_tests = parse_excluded_tests(
            os.path.join(new_source, 'test', 'api', 'serialized_plans', 'excluded_tests.list')
        )
        test_dir = os.path.join('test', 'sql')
        if new_source != '.':
            test_dir = os.path.join(new_source, test_dir)
        files = find_tests_recursive(test_dir, excluded_tests)
    elif args.test_list is not None:
        with open(args.test_list, 'r') as f:
            for line in f:
                if len(line.strip()) == 0:
                    continue
                files.append(line.strip())
    else:
        # run a single test
        files.append(args.test_file)
    files.sort()

    current_path = os.getcwd()
    try:
        build_sources(old_source, new_source)

        all_succeeded = True
        started = False
        if args.start_at is None:
            started = True
        for filename in files:
            if not started:
                if filename == args.start_at:
                    started = True
                else:
                    continue

            print(f"Run test {filename}")
            os.chdir(current_path)
            if not run_test(filename, old_source, new_source, args.no_exit):
                all_succeeded = False
        if not all_succeeded:
            exit(1)
    except:
        raise
    finally:
        os.chdir(current_path)


if __name__ == "__main__":
    main()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest


def test_analyze_truncation(shell):
    # if we find a long string after a bunch of short strings with a low analyze count, the string will get truncated
    test = (
        ShellTest(shell)
        .statement(".maxrows -1 1000")
        .statement("select i::varchar from range(10_000) t(i) union all select 'this is a very long string that will get truncated'")
    )
    result = test.run()
    result.check_stdout('9999')
    result.check_stdout('…')

def test_analyze_streaming_wrap(shell):
    # we can wrap also while streaming, if our max width is set
    test = (
        ShellTest(shell)
        .statement(".maxrows -1 100")
        .statement(".maxwidth 100")
        .statement("select concat('this is a long string that will wrap given the current maximum width that is set to only 100 which is not very wide (#', i, ')') from range(10000) t(i);")
    )
    result = test.run()
    result.check_stdout('#9999')


# fmt: on



import os
import csv
import re
import argparse
import glob
from typing import Set, Tuple, cast
import pathlib
from typing import NamedTuple
from typing import List, Dict
import json

os.chdir(os.path.join(os.path.dirname(__file__), '..'))

# Example usage:

parser = argparse.ArgumentParser(description='Generates/Validates extension_functions.hpp file')

parser.add_argument(
    '--validate',
    action=argparse.BooleanOptionalAction,
    help='If set will validate that extension_entries.hpp is up to date, otherwise it generates the extension_functions.hpp file.',
)
parser.add_argument(
    '--extension_repository',
    action='store',
    help="The repository to look for the '**/<extension>.duckdb_extension' files",
    default='build/release/repository',
)
parser.add_argument(
    '--shell',
    action='store',
    help="Path to the DuckDB shell",
    default='build/release/duckdb',
)
parser.add_argument(
    '--extensions',
    action='store',
    help="Comma separated list of extensions - if not provided this is read from the extension configuration",
    default='',
)

args = parser.parse_args()

EXTENSIONS_PATH = os.path.join("build", "extension_configuration", "extensions.csv")
DUCKDB_PATH = os.path.join(*args.shell.split('/'))
HEADER_PATH = os.path.join("src", "include", "duckdb", "main", "extension_entries.hpp")

EXTENSION_DEPENDENCIES = {
    'iceberg': [
        'avro',
        'parquet',
    ]
}

from enum import Enum


class CatalogType(str, Enum):
    SCALAR = "CatalogType::SCALAR_FUNCTION_ENTRY"
    TABLE = "CatalogType::TABLE_FUNCTION_ENTRY"
    AGGREGATE = "CatalogType::AGGREGATE_FUNCTION_ENTRY"
    PRAGMA = "CatalogType::PRAGMA_FUNCTION_ENTRY"
    MACRO = "CatalogType::MACRO_ENTRY"
    TABLE_MACRO = "CatalogType::TABLE_MACRO_ENTRY"


parameter_type_map = {"TIMESTAMP WITH TIME ZONE": "TIMESTAMPTZ", "TIME WITH TIME ZONE": "TIMETZ"}


def catalog_type_from_type(catalog_type: str) -> CatalogType:
    TYPE_MAP = {
        CatalogType.SCALAR.value: CatalogType.SCALAR,
        CatalogType.TABLE.value: CatalogType.TABLE,
        CatalogType.AGGREGATE.value: CatalogType.AGGREGATE,
        CatalogType.PRAGMA.value: CatalogType.PRAGMA,
        CatalogType.MACRO.value: CatalogType.MACRO,
        CatalogType.TABLE_MACRO.value: CatalogType.TABLE_MACRO,
    }
    if catalog_type not in TYPE_MAP:
        raise Exception(f"Unrecognized function type: '{catalog_type}'")
    return TYPE_MAP[catalog_type]


def catalog_type_from_string(catalog_type: str) -> CatalogType:
    TYPE_MAP = {
        CatalogType.SCALAR.name.lower(): CatalogType.SCALAR,
        CatalogType.TABLE.name.lower(): CatalogType.TABLE,
        CatalogType.AGGREGATE.name.lower(): CatalogType.AGGREGATE,
        CatalogType.PRAGMA.name.lower(): CatalogType.PRAGMA,
        CatalogType.MACRO.name.lower(): CatalogType.MACRO,
        CatalogType.TABLE_MACRO.name.lower(): CatalogType.TABLE_MACRO,
    }
    if catalog_type not in TYPE_MAP:
        raise Exception(f"Unrecognized function type: '{catalog_type}'")
    return TYPE_MAP[catalog_type]


def parse_records(text):
    records = []  # Will hold all parsed records
    current_record = []  # Holds items for the current record
    current_item = []  # Accumulates characters for the current item
    in_quote = False  # True if we're inside a double-quoted string
    inside_braces = False  # True if we're inside a { ... } block

    for char in text:
        if char == '"':
            # Toggle the quote state and always include the quote.
            in_quote = not in_quote
        elif char == '{' and not in_quote:
            # Start of a new record.
            inside_braces = True
            # Reset any previous record state.
            current_record = []
            current_item = []
        elif char == '}' and not in_quote and inside_braces:
            # End of the current record.
            token = ''.join(current_item).strip()
            if token:
                current_record.append(token)
            records.append(current_record)
            # Reset state for subsequent records.
            current_record = []
            current_item = []
            inside_braces = False
        elif char == ',' and not in_quote and inside_braces:
            # A comma outside quotes indicates the end of the current item.
            token = ''.join(current_item).strip()
            if token:
                current_record.append(token)
            current_item = []
        else:
            # Otherwise, just add the character if we're inside braces.
            if inside_braces:
                current_item.append(char)
    return records


class LogicalType(NamedTuple):
    type: str


class Function(NamedTuple):
    name: str
    type: CatalogType


class FunctionOverload(NamedTuple):
    name: str
    type: CatalogType
    parameters: Tuple
    return_type: LogicalType


class ExtensionFunctionOverload(NamedTuple):
    extension: str
    name: str
    type: CatalogType
    parameters: Tuple
    return_type: LogicalType

    @staticmethod
    def create_map(input: List[Tuple[str, str, str, str]]) -> Dict[Function, List["ExtensionFunctionOverload"]]:
        output: Dict[Function, List["ExtensionFunctionOverload"]] = {}
        for x in input:
            function = Function(x[0], catalog_type_from_type(x[2]))
            # parse the signature
            signature = x[3]
            splits = signature.split('>')
            return_type = LogicalType(splits[1])
            parameters = [LogicalType(param) for param in splits[0][1:-1].split(',')]
            extension_function = ExtensionFunctionOverload(x[1], function.name, function.type, parameters, return_type)
            if function not in output:
                output[function] = []
            output[function].append(extension_function)
        return output


class ExtensionFunction(NamedTuple):
    extension: str
    name: str
    type: CatalogType

    @staticmethod
    def create_map(input: List[Tuple[str, str, str]]) -> Dict[Function, "ExtensionFunction"]:
        output: Dict[Function, "ExtensionFunction"] = {}
        for x in input:
            key = Function(x[0], catalog_type_from_type(x[2]))
            output[key] = ExtensionFunction(x[1], key.name, key.type)
        return output


class ExtensionSetting(NamedTuple):
    extension: str
    name: str

    @staticmethod
    def create_map(input: List[Tuple[str, str]]) -> Dict[str, "ExtensionSetting"]:
        output: Dict[str, "ExtensionSetting"] = {}
        for x in input:
            output[x[0]] = ExtensionSetting(x[1], x[0])
        return output


class ExtensionSecretType(NamedTuple):
    extension: str
    name: str

    @staticmethod
    def create_map(input: List[Tuple[str, str]]) -> Dict[str, "ExtensionSecretType"]:
        output: Dict[str, "ExtensionSecretType"] = {}
        for x in input:
            output[x[0]] = ExtensionSecretType(x[1], x[0])
        return output


class ExtensionCopyFunction(NamedTuple):
    extension: str
    name: str

    @staticmethod
    def create_map(input: List[Tuple[str, str]]) -> Dict[str, "ExtensionCopyFunction"]:
        output: Dict[str, "ExtensionCopyFunction"] = {}
        for x in input:
            output[x[0]] = ExtensionCopyFunction(x[1], x[0])
        return output


class ExtensionType(NamedTuple):
    extension: str
    name: str

    @staticmethod
    def create_map(input: List[Tuple[str, str]]) -> Dict[str, "ExtensionType"]:
        output: Dict[str, "ExtensionType"] = {}
        for x in input:
            output[x[0]] = ExtensionType(x[1], x[0])
        return output


class ParsedEntries:
    def __init__(self, file_path):
        self.path = file_path
        self.functions = {}
        self.function_overloads = {}
        self.settings = {}
        self.secret_types = {}
        self.types = {}
        self.copy_functions = {}

        file = open(file_path, 'r')
        file_blob = file.read()

        # Get the extension functions
        ext_functions_file_blob = get_slice_of_file("EXTENSION_FUNCTIONS", file_blob)
        res = parse_records(ext_functions_file_blob)
        res = [(x[0], x[1], x[2]) for x in res]
        self.functions = ExtensionFunction.create_map(res)

        # Get the extension function overloads
        ext_function_overloads_file_blob = get_slice_of_file("EXTENSION_FUNCTION_OVERLOADS", file_blob)
        res = parse_records(ext_function_overloads_file_blob)
        res = [(x[0], x[1], x[2], x[3]) for x in res]
        self.function_overloads = ExtensionFunctionOverload.create_map(res)

        # Get the extension settings
        ext_settings_file_blob = get_slice_of_file("EXTENSION_SETTINGS", file_blob)
        res = parse_records(ext_settings_file_blob)
        res = [(x[0], x[1]) for x in res]
        self.settings = ExtensionSetting.create_map(res)

        # Get the extension secret types
        ext_secret_types_file_blob = get_slice_of_file("EXTENSION_SECRET_TYPES", file_blob)
        res = parse_records(ext_secret_types_file_blob)
        res = [(x[0], x[1]) for x in res]
        self.secret_types = ExtensionSecretType.create_map(res)

        # Get the extension types
        ext_copy_functions_blob = get_slice_of_file("EXTENSION_COPY_FUNCTIONS", file_blob)
        res = parse_records(ext_copy_functions_blob)
        res = [(x[0], x[1]) for x in res]
        self.copy_functions = ExtensionCopyFunction.create_map(res)

        # Get the extension types
        ext_types_file_blob = get_slice_of_file("EXTENSION_TYPES", file_blob)
        res = parse_records(ext_types_file_blob)
        res = [(x[0], x[1]) for x in res]
        self.types = ExtensionType.create_map(res)

    def strip_unloaded_extensions(self, extensions: List[str], functions):
        return [x for x in functions if x.extension not in extensions]

    def filter_entries(self, extensions: List[str]):
        self.functions = {k: v for k, v in self.functions.items() if v.extension not in extensions}
        self.function_overloads = {
            k: self.strip_unloaded_extensions(extensions, v)
            for k, v in self.function_overloads.items()
            if len(self.strip_unloaded_extensions(extensions, v)) > 0
        }
        self.copy_functions = {k: v for k, v in self.copy_functions.items() if v.extension not in extensions}
        self.settings = {k: v for k, v in self.settings.items() if v.extension not in extensions}
        self.secret_types = {k: v for k, v in self.secret_types.items() if v.extension not in extensions}
        self.types = {k: v for k, v in self.types.items() if v.extension not in extensions}


def check_prerequisites():
    if not os.path.isfile(DUCKDB_PATH):
        print(f"{DUCKDB_PATH} not found")
        print(
            "please run 'GENERATE_EXTENSION_ENTRIES=1 BUILD_ALL_EXT=1 make release', you might have to manually add DONT_LINK to all extension_configs"
        )
        exit(1)
    if len(args.extensions) == 0 and not os.path.isfile(EXTENSIONS_PATH):
        print(f"{EXTENSIONS_PATH} not found and --extensions it not set")
        print("Either:")
        print(
            "* run 'GENERATE_EXTENSION_ENTRIES=1 BUILD_ALL_EXT=1 make release', you might have to manually add DONT_LINK to all extension_configs"
        )
        print("* Specify a comma separated list of extensions using --extensions")
        exit(1)
    if not os.path.isdir(args.extension_repository):
        print(f"provided --extension_repository '{args.extension_repository}' is not a valid directory")
        exit(1)


# Parses the extension config files for which extension names there are to be expected
def get_extension_names() -> List[str]:
    if len(args.extensions) > 0:
        return args.extensions.split(',')
    extension_names = []
    with open(EXTENSIONS_PATH) as f:
        # Skip the csv header
        next(f)
        for line in f:
            extension_name = line.split(',')[0].rstrip()
            extension_names.append(extension_name)
    return extension_names


def get_query(sql_query, load_query) -> list:
    # Optionally perform a LOAD of an extension
    # Then perform a SQL query, fetch the output
    import csv
    import io

    query = f'{DUCKDB_PATH} -unsigned -csv -c "{load_query}{sql_query}" '
    query_result = os.popen(query).read()
    f = io.StringIO(query_result)
    reader = csv.reader(f)
    header = next(reader)
    result = []
    for line in reader:
        result_obj = {}
        for i in range(len(header)):
            result_obj[header[i]] = line[i]
        result.append(result_obj)
    return result


def transform_parameter(parameter) -> LogicalType:
    if parameter is None:
        return LogicalType("INVALID")
    parameter = parameter.upper()
    if parameter.endswith('[]'):
        return LogicalType(transform_parameter(parameter[0 : len(parameter) - 2]).type + '[]')
    if parameter in parameter_type_map:
        return LogicalType(parameter_type_map[parameter])
    return LogicalType(parameter)


def transform_parameters(parameters) -> FunctionOverload:
    parameters = [x for x in parameters.lstrip('[').rstrip(']').split(', ') if len(x) > 0]
    return tuple(transform_parameter(param) for param in parameters)


def get_functions(load="") -> (Set[Function], Dict[Function, List[FunctionOverload]]):
    GET_FUNCTIONS_QUERY = """
        select distinct
            function_name,
            function_type,
            parameter_types,
            return_type
        from duckdb_functions()
        ORDER BY function_name, function_type;
    """
    # ['name_1,type_1', ..., 'name_n,type_n']
    results = get_query(GET_FUNCTIONS_QUERY, load)

    functions = set()
    function_overloads = {}
    for func in results:
        function_name = func["function_name"].lower()
        function_type = func["function_type"]
        parameter_types = func["parameter_types"]
        return_type = func["return_type"]
        function_parameters = transform_parameters(parameter_types)
        function_return = transform_parameter(return_type)
        function = Function(function_name, catalog_type_from_string(function_type))
        function_overload = FunctionOverload(
            function_name, catalog_type_from_string(function_type), function_parameters, function_return
        )
        if function not in functions:
            functions.add(function)
            function_overloads[function] = [function_overload]
        else:
            function_overloads[function].append(function_overload)

    return (functions, function_overloads)


def get_settings(load="") -> Set[str]:
    GET_SETTINGS_QUERY = """
        select distinct
            name
        from duckdb_settings();
    """
    settings = get_query(GET_SETTINGS_QUERY, load)
    res = set()
    for setting in settings:
        name = setting['name']
        res.add(name)
    return res


def get_secret_types(load="") -> Set[str]:
    GET_SECRET_TYPES_QUERY = """
        select distinct
            type
        from duckdb_secret_types();
    """
    secret_types = get_query(GET_SECRET_TYPES_QUERY, load)
    res = set()
    for secret_type in secret_types:
        type = secret_type['type']
        res.add(type)
    return res


class ExtensionData:
    def __init__(self):
        # Map of extension -> ExtensionFunction
        self.function_map: Dict[Function, ExtensionFunction] = {}
        # Map of extension -> ExtensionSetting
        self.settings_map: Dict[str, ExtensionSetting] = {}
        # Map of extension -> ExtensionSecretType
        self.secret_types_map: Dict[str, ExtensionSecretType] = {}
        # Map of function -> extension function overloads
        self.function_overloads: Dict[Function, List[ExtensionFunctionOverload]] = {}
        # All function overloads (also ones that will not be written to the file)
        self.all_function_overloads: Dict[Function, List[ExtensionFunctionOverload]] = {}

        self.base_settings: Set[str] = set()
        self.base_secret_types: Set[str] = set()
        self.base_functions: Set[Function] = set()

        self.extension_settings: Dict[str, Set[str]] = {}
        self.extension_secret_types: Dict[str, Set[str]] = {}
        self.extension_functions: Dict[str, Set[Function]] = {}

        self.added_extensions: Set[str] = set()

        # Map of extension -> extension_path
        self.extensions: Dict[str, str] = get_extension_path_map()

        self.stored_functions: Dict[str, List[Function]] = {
            'arrow': [Function("scan_arrow_ipc", CatalogType.TABLE), Function("to_arrow_ipc", CatalogType.TABLE)],
            'spatial': [],
        }
        self.stored_settings: Dict[str, List[str]] = {'arrow': [], 'spatial': []}

    def set_base(self):
        (functions, function_overloads) = get_functions()
        self.base_functions: Set[Function] = functions
        self.base_settings: Set[str] = get_settings()
        self.base_secret_types: Set[str] = get_secret_types()

    def add_entries(self, entries: ParsedEntries):
        self.function_map.update(entries.functions)
        self.function_overloads.update(entries.function_overloads)
        self.settings_map.update(entries.settings)
        self.secret_types_map.update(entries.secret_types)

    def load_dependencies(self, extension_name: str) -> str:
        if extension_name not in EXTENSION_DEPENDENCIES:
            return ''

        res = ''
        dependencies = EXTENSION_DEPENDENCIES[extension_name]
        for item in dependencies:
            if item not in self.extensions:
                print(f"Could not load extension '{extension_name}', dependency '{item}' is missing")
                exit(1)
            extension_path = self.extensions[item]
            print(f"Load {item} at {extension_path}")
            res += f"LOAD '{extension_path}';"
        return res

    def add_extension(self, extension_name: str):
        if extension_name in EXTENSION_DEPENDENCIES:
            for item in EXTENSION_DEPENDENCIES[extension_name]:
                if item not in self.added_extensions:
                    self.add_extension(item)

        if extension_name in self.extensions:
            # Perform a LOAD and add the added settings/functions/secret_types
            extension_path = self.extensions[extension_name]

            print(f"Load {extension_name} at {extension_path}")
            load = self.load_dependencies(extension_name)
            load += f"LOAD '{extension_path}';"

            (functions, function_overloads) = get_functions(load)
            extension_functions = list(functions)
            extension_settings = list(get_settings(load))
            extension_secret_types = list(get_secret_types(load))

            self.add_settings(extension_name, extension_settings)
            self.add_secret_types(extension_name, extension_secret_types)
            self.add_functions(extension_name, extension_functions, function_overloads)
        elif extension_name in self.stored_functions or extension_name in self.stored_settings:
            # Retrieve the list of settings/functions from our hardcoded list
            extension_functions = self.stored_functions[extension_name]
            extension_settings = self.stored_settings[extension_name]
            extension_secret_types = self.stored_secret_types[extension_name]

            print(f"Loading {extension_name} from stored functions: {extension_functions}")
            self.add_settings(extension_name, extension_settings)
            self.add_secret_types(extension_name, extension_secret_types)
            self.add_functions(extension_name, extension_functions)
        else:
            error = f"""Missing extension {extension_name} and not found in stored_functions/stored_settings/stored_secret_types
Please double check if '{args.extension_repository}' is the right location to look for ./**/*.duckdb_extension files"""
            print(error)
            exit(1)
        self.added_extensions.add(extension_name)

    def add_settings(self, extension_name: str, settings_list: List[str]):
        extension_name = extension_name.lower()

        base_settings = set()
        base_settings.update(self.base_settings)
        if extension_name in EXTENSION_DEPENDENCIES:
            dependencies = EXTENSION_DEPENDENCIES[extension_name]
            for item in dependencies:
                assert item in self.extension_settings
                base_settings.update(self.extension_settings[item])

        added_settings: Set[str] = set(settings_list) - base_settings

        self.extension_settings[extension_name] = added_settings

        settings_to_add: Dict[str, ExtensionSetting] = {}
        for setting in added_settings:
            setting_name = setting.lower()
            settings_to_add[setting_name] = ExtensionSetting(extension_name, setting_name)

        self.settings_map.update(settings_to_add)

    def add_secret_types(self, extension_name: str, secret_types_list: List[str]):
        extension_name = extension_name.lower()

        base_secret_types = set()
        base_secret_types.update(self.base_secret_types)
        if extension_name in EXTENSION_DEPENDENCIES:
            dependencies = EXTENSION_DEPENDENCIES[extension_name]
            for item in dependencies:
                assert item in self.extension_secret_types
                base_secret_types.update(self.extension_secret_types[item])

        added_secret_types: Set[str] = set(secret_types_list) - base_secret_types

        self.extension_secret_types[extension_name] = added_secret_types

        secret_types_to_add: Dict[str, ExtensionSecretType] = {}
        for secret_type in added_secret_types:
            secret_type_name = secret_type.lower()
            secret_types_to_add[secret_type_name] = ExtensionSecretType(extension_name, secret_type_name)

        self.secret_types_map.update(secret_types_to_add)

    def get_extension_overloads(
        self, extension_name: str, overloads: Dict[Function, List[FunctionOverload]]
    ) -> Dict[Function, List[ExtensionFunctionOverload]]:
        result = {}
        for function, function_overloads in overloads.items():
            extension_overloads = []
            for overload in function_overloads:
                extension_overloads.append(
                    ExtensionFunctionOverload(
                        extension_name, overload.name, overload.type, overload.parameters, overload.return_type
                    )
                )
            result[function] = extension_overloads
        return result

    def add_functions(
        self, extension_name: str, function_list: List[Function], overloads: Dict[Function, List[FunctionOverload]]
    ):
        extension_name = extension_name.lower()

        base_functions = set()
        base_functions.update(self.base_functions)
        if extension_name in EXTENSION_DEPENDENCIES:
            dependencies = EXTENSION_DEPENDENCIES[extension_name]
            for item in dependencies:
                assert item in self.extension_functions
                base_functions.update(self.extension_functions[item])

        overloads = self.get_extension_overloads(extension_name, overloads)
        added_functions: Set[Function] = set(function_list) - base_functions

        self.extension_functions[extension_name] = added_functions

        functions_to_add: Dict[Function, ExtensionFunction] = {}
        for function in added_functions:
            if function in self.function_overloads:
                # function is in overload map - add overloads
                self.function_overloads[function] += overloads[function]
            elif function in self.function_map:
                # function is in function map and we are trying to add it again
                # this means the function is present in multiple extensions
                # remove from function map, and add to overload map
                self.function_overloads[function] = self.all_function_overloads[function] + overloads[function]
                del self.function_map[function]
            else:
                functions_to_add[function] = ExtensionFunction(extension_name, function.name, function.type)

        self.all_function_overloads.update(overloads)
        self.function_map.update(functions_to_add)

    def validate(self):
        parsed_entries = ParsedEntries(HEADER_PATH)
        if self.function_map != parsed_entries.functions:
            print("Function map mismatches:")
            print_map_diff(self.function_map, parsed_entries.functions)
            exit(1)
        if self.settings_map != parsed_entries.settings:
            print("Settings map mismatches:")
            print_map_diff(self.settings_map, parsed_entries.settings)
            exit(1)
        if self.secret_types_map != parsed_entries.secret_types:
            print("SecretTypes map mismatches:")
            print_map_diff(self.secret_types_map, parsed_entries.secret_types)
            exit(1)

        print("All entries found: ")
        print(" > functions: " + str(len(parsed_entries.functions)))
        print(" > settings:  " + str(len(parsed_entries.settings)))
        print(" > secret_types:  " + str(len(parsed_entries.secret_types)))

    def verify_export(self):
        if len(self.function_map) == 0 or len(self.settings_map) == 0 or len(self.secret_types_map) == 0:
            print(
                """
The provided configuration produced an empty function map or empty settings map or empty secret types map
This is likely caused by building DuckDB with extensions linked in
"""
            )
            exit(1)

    def export_functions(self) -> str:
        result = """
static constexpr ExtensionFunctionEntry EXTENSION_FUNCTIONS[] = {\n"""
        sorted_function = sorted(self.function_map)

        for func in sorted_function:
            function: ExtensionFunction = self.function_map[func]
            result += "\t{"
            result += f'"{function.name}", "{function.extension}", {function.type.value}'
            result += "},\n"
        result += "}; // END_OF_EXTENSION_FUNCTIONS\n"
        return result

    def export_function_overloads(self) -> str:
        result = """
static constexpr ExtensionFunctionOverloadEntry EXTENSION_FUNCTION_OVERLOADS[] = {\n"""
        sorted_function = sorted(self.function_overloads)

        for func in sorted_function:
            overloads: List[ExtensionFunctionOverload] = sorted(self.function_overloads[func])
            for overload in overloads:
                result += "\t{"
                result += f'"{overload.name}", "{overload.extension}", {overload.type.value}, "'
                signature = "["
                signature += ",".join([parameter.type for parameter in overload.parameters])
                signature += "]>" + overload.return_type.type
                result += signature
                result += '"},\n'
        result += "}; // END_OF_EXTENSION_FUNCTION_OVERLOADS\n"
        return result

    def export_settings(self) -> str:
        result = """
static constexpr ExtensionEntry EXTENSION_SETTINGS[] = {\n"""
        sorted_settings = sorted(self.settings_map)

        for settings_name in sorted_settings:
            setting: ExtensionSetting = self.settings_map[settings_name]
            result += "\t{"
            result += f'"{settings_name.lower()}", "{setting.extension}"'
            result += "},\n"
        result += "}; // END_OF_EXTENSION_SETTINGS\n"
        return result

    def export_secret_types(self) -> str:
        result = """
static constexpr ExtensionEntry EXTENSION_SECRET_TYPES[] = {\n"""
        sorted_secret_types = sorted(self.secret_types_map)

        for secret_types_name in sorted_secret_types:
            secret_type: ExtensionSecretType = self.secret_types_map[secret_types_name]
            result += "\t{"
            result += f'"{secret_types_name.lower()}", "{secret_type.extension}"'
            result += "},\n"
        result += "}; // END_OF_EXTENSION_SECRET_TYPES\n"
        return result


# Get the slice of the file containing the var (assumes // END_OF_<varname> comment after var)
def get_slice_of_file(var_name, file_str):
    begin = file_str.find(var_name)
    end = file_str.find("END_OF_" + var_name)
    return file_str[begin:end]


def print_map_diff(d1, d2):
    s1 = sorted(set(d1.items()))
    s2 = sorted(set(d2.items()))

    diff1 = str(set(s1) - set(s2))
    diff2 = str(set(s2) - set(s1))
    print("Diff between maps: " + diff1 + "\n")
    print("Diff between maps: " + diff2 + "\n")


def get_extension_path_map() -> Dict[str, str]:
    extension_paths: Dict[str, str] = {}
    # extension_repository = pathlib.Path('../build/release/repository')
    extension_repository = args.extension_repository
    for location in glob.iglob(extension_repository + '/**/*.duckdb_extension', recursive=True):
        name, _ = os.path.splitext(os.path.basename(location))
        print(f"Located extension: {name} in path: '{location}'")
        extension_paths[name] = location
    return extension_paths


def write_header(data: ExtensionData):
    INCLUDE_HEADER = """//===----------------------------------------------------------------------===//
//                         DuckDB
//
// duckdb/main/extension_entries.hpp
//
//
//===----------------------------------------------------------------------===//

#pragma once

#include \"duckdb/common/unordered_map.hpp\"
#include \"duckdb/common/enums/catalog_type.hpp\"

// NOTE: this file is generated by scripts/generate_extensions_function.py.
// Example usage to refresh one extension (replace "icu" with the desired extension):
// GENERATE_EXTENSION_ENTRIES=1 make debug
// python3 scripts/generate_extensions_function.py --extensions icu --shell build/debug/duckdb --extension_repository build/debug/repository

// Check out the check-load-install-extensions  job in .github/workflows/LinuxRelease.yml for more details

namespace duckdb {

struct ExtensionEntry {
    char name[48];
    char extension[48];
};

struct ExtensionFunctionEntry {
    char name[48];
    char extension[48];
    CatalogType type;
};

struct ExtensionFunctionOverloadEntry {
    char name[48];
    char extension[48];
    CatalogType type;
    char signature[96];
};
"""

    INCLUDE_FOOTER = """
// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_COPY_FUNCTIONS[] = {
    {"parquet", "parquet"},
    {"json", "json"},
    {"avro", "avro"},
    {"iceberg", "iceberg"}
}; // END_OF_EXTENSION_COPY_FUNCTIONS

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_TYPES[] = {
    {"json", "json"},
    {"inet", "inet"},
}; // END_OF_EXTENSION_TYPES

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_COLLATIONS[] = {
    {"af", "icu"},    {"am", "icu"},    {"ar", "icu"},     {"ar_sa", "icu"}, {"as", "icu"},    {"az", "icu"},
    {"be", "icu"},    {"bg", "icu"},    {"bn", "icu"},     {"bo", "icu"},    {"br", "icu"},    {"bs", "icu"},
    {"ca", "icu"},    {"ceb", "icu"},   {"chr", "icu"},    {"cs", "icu"},    {"cy", "icu"},    {"da", "icu"},
    {"de", "icu"},    {"de_at", "icu"}, {"dsb", "icu"},    {"dz", "icu"},    {"ee", "icu"},    {"el", "icu"},
    {"en", "icu"},    {"en_us", "icu"}, {"eo", "icu"},     {"es", "icu"},    {"et", "icu"},    {"fa", "icu"},
    {"fa_af", "icu"}, {"ff", "icu"},    {"fi", "icu"},     {"fil", "icu"},   {"fo", "icu"},    {"fr", "icu"},
    {"fr_ca", "icu"}, {"fy", "icu"},    {"ga", "icu"},     {"gl", "icu"},    {"gu", "icu"},    {"ha", "icu"},
    {"haw", "icu"},   {"he", "icu"},    {"he_il", "icu"},  {"hi", "icu"},    {"hr", "icu"},    {"hsb", "icu"},
    {"hu", "icu"},    {"hy", "icu"},    {"id", "icu"},     {"id_id", "icu"}, {"ig", "icu"},    {"is", "icu"},
    {"it", "icu"},    {"ja", "icu"},    {"ka", "icu"},     {"kk", "icu"},    {"kl", "icu"},    {"km", "icu"},
    {"kn", "icu"},    {"ko", "icu"},    {"kok", "icu"},    {"ku", "icu"},    {"ky", "icu"},    {"lb", "icu"},
    {"lkt", "icu"},   {"ln", "icu"},    {"lo", "icu"},     {"lt", "icu"},    {"lv", "icu"},    {"mk", "icu"},
    {"ml", "icu"},    {"mn", "icu"},    {"mr", "icu"},     {"ms", "icu"},    {"mt", "icu"},    {"my", "icu"},
    {"nb", "icu"},    {"nb_no", "icu"}, {"ne", "icu"},     {"nl", "icu"},    {"nn", "icu"},    {"om", "icu"},
    {"or", "icu"},    {"pa", "icu"},    {"pa_in", "icu"},  {"pl", "icu"},    {"ps", "icu"},    {"pt", "icu"},
    {"ro", "icu"},    {"ru", "icu"},    {"sa", "icu"},     {"se", "icu"},    {"si", "icu"},    {"sk", "icu"},
    {"sl", "icu"},    {"smn", "icu"},   {"sq", "icu"},     {"sr", "icu"},    {"sr_ba", "icu"}, {"sr_me", "icu"},
    {"sr_rs", "icu"}, {"sv", "icu"},    {"sw", "icu"},     {"ta", "icu"},    {"te", "icu"},    {"th", "icu"},
    {"tk", "icu"},    {"to", "icu"},    {"tr", "icu"},     {"ug", "icu"},    {"uk", "icu"},    {"ur", "icu"},
    {"uz", "icu"},    {"vi", "icu"},    {"wae", "icu"},    {"wo", "icu"},    {"xh", "icu"},    {"yi", "icu"},
    {"yo", "icu"},    {"yue", "icu"},   {"yue_cn", "icu"}, {"zh", "icu"},    {"zh_cn", "icu"}, {"zh_hk", "icu"},
    {"zh_mo", "icu"}, {"zh_sg", "icu"}, {"zh_tw", "icu"},  {"zu", "icu"}}; // END_OF_EXTENSION_COLLATIONS

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_FILE_PREFIXES[] = {
     {"http://", "httpfs"}, {"https://", "httpfs"}, {"s3://", "httpfs"}, {"s3a://", "httpfs"}, {"s3n://", "httpfs"},
     {"gcs://", "httpfs"},  {"gs://", "httpfs"},    {"r2://", "httpfs"}, {"azure://", "azure"}, {"az://", "azure"},
     {"abfss://", "azure"}, {"hf://", "httpfs"}
}; // END_OF_EXTENSION_FILE_PREFIXES

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_FILE_POSTFIXES[] = {
    {".parquet", "parquet"},
    {".json", "json"},
    {".jsonl", "json"},
    {".ndjson", "json"},
    {".shp", "spatial"},
    {".gpkg", "spatial"},
    {".fgb", "spatial"},
    {".xlsx", "excel"},
    {".avro", "avro"},
}; // END_OF_EXTENSION_FILE_POSTFIXES

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_FILE_CONTAINS[] = {
    {".parquet?", "parquet"},
    {".json?", "json"},
    {".ndjson?", ".jsonl?"},
    {".jsonl?", ".ndjson?"}
}; // EXTENSION_FILE_CONTAINS

// Note: these are currently hardcoded in scripts/generate_extensions_function.py
// TODO: automate by passing though to script via duckdb
static constexpr ExtensionEntry EXTENSION_SECRET_PROVIDERS[] = {{"s3/config", "httpfs"},
                                                                {"gcs/config", "httpfs"},
                                                                {"r2/config", "httpfs"},
                                                                {"s3/credential_chain", "aws"},
                                                                {"gcs/credential_chain", "aws"},
                                                                {"r2/credential_chain", "aws"},
                                                                {"aws/credential_chain", "aws"},
                                                                {"rds/credential_chain", "aws"},
                                                                {"azure/access_token", "azure"},
                                                                {"azure/config", "azure"},
                                                                {"azure/credential_chain", "azure"},
                                                                {"azure/service_principal", "azure"},
                                                                {"huggingface/config", "httfps"},
                                                                {"huggingface/credential_chain", "httpfs"},
                                                                {"bearer/config", "httpfs"},
                                                                {"mysql/config", "mysql_scanner"},
                                                                {"postgres/config", "postgres_scanner"}
}; // EXTENSION_SECRET_PROVIDERS

static constexpr const char *AUTOLOADABLE_EXTENSIONS[] = {
    "autocomplete",
    "avro",
    "aws",
    "azure",
    "core_functions",
    "delta",
    "ducklake",
    "encodings",
    "excel",
    "fts",
    "httpfs",
    "iceberg",
    "icu",
    "inet",
    "json",
    "motherduck",
    "mysql_scanner",
    "parquet",
    "postgres_scanner",
    "quack",
    "sqlite_scanner",
    "sqlsmith",
    "tpcds",
    "tpch",
    "ui",
    "unity_catalog"
}; // END_OF_AUTOLOADABLE_EXTENSIONS

} // namespace duckdb"""

    data.verify_export()

    file = open(HEADER_PATH, 'w')
    file.write(INCLUDE_HEADER)

    exported_functions = data.export_functions()
    file.write(exported_functions)

    exported_overloads = data.export_function_overloads()
    file.write(exported_overloads)

    exported_settings = data.export_settings()
    file.write(exported_settings)

    exported_secret_types = data.export_secret_types()
    file.write(exported_secret_types)

    file.write(INCLUDE_FOOTER)
    file.close()


# Extensions that can be autoloaded, but are not buildable by DuckDB CI
HARDCODED_EXTENSION_FUNCTIONS = ExtensionFunction.create_map(
    [
        ("delta_scan", "delta", "CatalogType::TABLE_FUNCTION_ENTRY"),
    ]
)


def main():
    check_prerequisites()

    extension_names: List[str] = get_extension_names()

    extension_data = ExtensionData()
    # Collect the list of functions/settings without any extensions loaded
    extension_data.set_base()

    # TODO: add 'purge' option to ignore existing entries ??
    parsed_entries = ParsedEntries(HEADER_PATH)
    parsed_entries.filter_entries(extension_names)

    # Add the entries we parsed from the HEADER_PATH
    extension_data.add_entries(parsed_entries)

    for extension_name in extension_names:
        print(extension_name)
        # For every extension, add the functions/settings added by the extension
        extension_data.add_extension(extension_name)

    # Add hardcoded extension entries (
    for key, value in HARDCODED_EXTENSION_FUNCTIONS.items():
        extension_data.function_map[key] = value

    if args.validate:
        extension_data.validate()
        return

    write_header(extension_data)


if __name__ == '__main__':
    main()



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
from tools.shell.tests.conftest import random_filepath


@pytest.mark.parametrize("command", [".sh ls", ".cd ..", ".log file", ".import file.csv tbl", ".open new_file", ".output out", ".once out", ".excel out", ".read myfile.sql"])
def test_safe_mode_command(shell, command):
    test = (
        ShellTest(shell, ['-safe'])
        .statement(command)
    )
    result = test.run()
    result.check_stderr('cannot be used in -safe mode')


@pytest.mark.parametrize("param", [(".sh ls", 'cannot be used in -safe mode'), ("INSTALL extension", "Permission Error")])
def test_safe_mode_dot_command(shell, param):
    command = param[0]
    expected_error = param[1]
    test = (
        ShellTest(shell)
        .statement('.safe_mode')
        .statement(command)
    )
    result = test.run()
    result.check_stderr(expected_error)

def test_safe_mode_database_basic(shell, random_filepath):
    test = (
        ShellTest(shell, [random_filepath, '-safe'])
        .statement('CREATE TABLE integers(i INT)')
        .statement('INSERT INTO integers VALUES (1), (2), (3)')
        .statement('SELECT SUM(i) FROM integers')
    )
    result = test.run()
    result.check_stdout("6")

@pytest.mark.parametrize("command", [".sh ls", ".cd ..", ".log file", ".import file.csv tbl", ".open new_file", ".output out", ".once out", ".excel out", ".read myfile.sql"])
@pytest.mark.parametrize("persistent", [False, True])
def test_safe_mode_database_commands(shell, random_filepath, command, persistent):
    arguments = ['-safe'] if not persistent else [random_filepath, '-safe']
    test = (
        ShellTest(shell, arguments)
        .statement(command)
    )
    result = test.run()
    result.check_stderr('cannot be used in -safe mode')

@pytest.mark.parametrize("sql", ["COPY (SELECT 42) TO 'test.csv'", "LOAD spatial", "INSTALL spatial", "ATTACH 'file.db' AS file"])
@pytest.mark.parametrize("persistent", [False, True])
def test_safe_mode_query(shell, random_filepath, sql, persistent):
    arguments = ['-safe'] if not persistent else [random_filepath, '-safe']
    test = (
        ShellTest(shell, arguments)
        .statement(sql)
    )
    result = test.run()
    result.check_stderr('disabled')


def test_lock_config(shell, random_filepath):
    arguments = ['-safe']
    test = (
        ShellTest(shell, arguments)
        .statement("SET memory_limit='-1'")
    )
    result = test.run()
    result.check_stderr('locked')

# fmt: on



import subprocess
import duckdb
import os
import pandas as pd
import argparse
from io import StringIO

parser = argparse.ArgumentParser(description='Rerun failed workflows from a PR.')
parser.add_argument(
    '--title',
    dest='title',
    action='store',
    help='The title of the PR for which we want to rerun workflows (or part of the title)',
    required=True,
)
parser.add_argument(
    '--repo', dest='repo', action='store', help='The repository to run this workflow on', default='duckdb/duckdb'
)
parser.add_argument(
    '--max_workflows',
    dest='max_workflows',
    action='store',
    help='The maximum number of workflows to look at (starting from the latest)',
    default=200,
)
args = parser.parse_args()

nlimit = args.max_workflows
query = args.title


proc = subprocess.Popen(
    [
        'gh',
        'run',
        '-R',
        args.repo,
        'list',
        '--json',
        'displayTitle,databaseId,status,conclusion,headSha',
        f'--limit={nlimit}',
    ],
    stdout=subprocess.PIPE,
)
text = proc.stdout.read().decode('utf8')
df = pd.read_json(StringIO(text))
result = duckdb.query(f"select headSha from df where displayTitle LIKE '%{query}%' limit 1").fetchall()
if len(result) == 0:
    print(
        f"No workflows found in the latest {nlimit} workflows that contain the text {query}.\nPerhaps try running with a higher --max_workflows parameter?"
    )
    exit(1)

headSha = result[0][0]

result = duckdb.query(
    f"select databaseId from df where conclusion IN ('failure', 'cancelled') AND displayTitle LIKE '%{query}%' and headSha='{headSha}'"
).fetchall()
if len(result) == 0:
    print(f"Found runs that match the text {query} but no failing or cancelled runs were found")
for databaseId in [x[0] for x in result]:
    os.system(f'gh run -R {args.repo} rerun {databaseId}')



from __future__ import annotations
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import Dict, List, Tuple, Optional, Iterable

from .inputs import validate_identifier


@dataclass(frozen=True)
class MetricDef:
    name: str
    group: str  # e.g., "file", "default", ...
    type: str  # e.g., "double", "uint64", "string", "map", "uint8"
    query_root: bool = False
    is_default: bool = False
    collection_method: Optional[str] = None
    child: Optional[str] = None
    description: str = None
    enum_value: Optional[int] = None


class MetricIndex:
    # Values 26-90 are reserved for optimizer metrics (auto-assigned sequentially).
    # Non-optimizer metrics must use values in [0, 25] or [91, 255].
    OPTIMIZER_RANGE_START = 26
    OPTIMIZER_RANGE_END = 90

    def __init__(self, defs: Iterable[MetricDef], optimizers: List[Tuple[str, int]]):
        self.defs: List[MetricDef] = list(defs)

        # Build name -> explicit value mapping
        self._value_of: Dict[str, int] = {}
        for d in self.defs:
            if d.enum_value is not None:
                self._value_of[d.name] = d.enum_value

        # Build group → names (existing contract for emitters)
        by_group: Dict[str, List[str]] = defaultdict(list)
        for d in self.defs:
            by_group[d.group].append(d.name)
        for g in by_group:
            by_group[g].sort()

        # Add optimizer group — derive MetricType values from OptimizerType numeric values
        if not optimizers:
            raise ValueError("No optimizers found in optimizer_type.hpp")
        # OptimizerType values start at 1 (EXPRESSION_REWRITER); INVALID = 0 is excluded.
        first_opt_value = 1
        optimizer_names = []
        for opt_name, opt_value in optimizers:
            metric_name = f"OPTIMIZER_{opt_name}"
            metric_value = self.OPTIMIZER_RANGE_START + (opt_value - first_opt_value)
            self._value_of[metric_name] = metric_value
            optimizer_names.append(metric_name)
        by_group["optimizer"] = optimizer_names
        value_to_name: Dict[int, str] = {}
        for name, val in self._value_of.items():
            if val < 0 or val > 255:
                raise ValueError(f"MetricType value {val} for '{name}' is outside uint8_t range (0-255)")
            if name in value_to_name.values():
                continue
            if val in value_to_name:
                raise ValueError(f"Duplicate MetricType value {val}: '{value_to_name[val]}' and '{name}'")
            is_optimizer = name.startswith("OPTIMIZER_")
            if is_optimizer and not (self.OPTIMIZER_RANGE_START <= val <= self.OPTIMIZER_RANGE_END):
                raise ValueError(
                    f"Optimizer metric '{name}' has value {val}, "
                    f"must be in [{self.OPTIMIZER_RANGE_START}, {self.OPTIMIZER_RANGE_END}]"
                )
            if not is_optimizer and self.OPTIMIZER_RANGE_START <= val <= self.OPTIMIZER_RANGE_END:
                raise ValueError(
                    f"Non-optimizer metric '{name}' has value {val}, "
                    f"which is in the reserved optimizer range [{self.OPTIMIZER_RANGE_START}, {self.OPTIMIZER_RANGE_END}]. "
                    f"New non-optimizer metrics should use values >= {self.OPTIMIZER_RANGE_END + 1}."
                )
            value_to_name[val] = name

        # Add "all"
        all_set = set().union(*by_group.values()) if by_group else set()
        by_group["all"] = sorted(all_set)

        # Add "default" based on "is_default" flag
        self.is_default_metrics: List[str] = []
        for d in self.defs:
            if d.is_default:
                self.is_default_metrics.append(d.name)
        self.is_default_metrics.sort()
        by_group["default"] = self.is_default_metrics

        # Deterministic order of groups
        self.metrics_by_group: Dict[str, List[str]] = OrderedDict(sorted(by_group.items()))

        # List of group names from metrics_by_group
        self.group_names: List[str] = sorted(list(self.metrics_by_group.keys()))

        # Root scope list
        self.root_scope: List[str] = sorted({d.name for d in self.defs if d.query_root})

        # Type → names (for profiling utils)
        type_map: Dict[str, List[str]] = defaultdict(list)
        for d in self.defs:
            t = d.type if d.name != "OPERATOR_TYPE" else "uint8"  # preserve your special-case
            type_map[t].append(d.name)
        for t in type_map:
            type_map[t].sort()
        self.type_to_names: Dict[str, List[str]] = dict(type_map)

        # Collection method → names (only for those that have it)
        coll_map: Dict[str, List[str]] = defaultdict(list)
        for d in self.defs:
            if d.collection_method:
                coll_map[d.collection_method].append(d.name)
        for c in coll_map:
            coll_map[c].sort()
        self.collection_to_names: Dict[str, List[str]] = dict(coll_map)

        # Children mapping (name → child-key)
        self.child_of: Dict[str, str] = {d.name: d.child for d in self.defs if d.child is not None}

    def all_metrics(self) -> List[str]:
        return self.metrics_by_group["all"]

    def metrics_per_group(self, group: str) -> List[str]:
        return self.metrics_by_group[group]

    def root_scope_metrics(self) -> List[str]:
        return self.root_scope

    def types_index(self) -> Dict[str, List[str]]:
        return self.type_to_names

    def collection_index(self) -> Dict[str, List[str]]:
        return self.collection_to_names

    def metrics_per_collection(self, collection: str) -> List[str]:
        return self.collection_to_names[collection]

    def child_index(self) -> Dict[str, str]:
        return self.child_of

    def metric_type(self, metric: str) -> str:
        for d in self.defs:
            if d.name == metric:
                if d.type == "uint8":
                    return "uint8_t"
                elif d.type == "uint64":
                    return "uint64_t"
                return d.type
        raise Exception(f"Unknown metric: {metric}")

    def metric_child(self, metric: str) -> Optional[str]:
        for d in self.defs:
            if d.name == metric:
                return d.child
        return None

    def metric_description(self, metric: str) -> Optional[str]:
        for d in self.defs:
            if d.name == metric:
                return d.description
        return None

    def metric_value(self, metric: str) -> Optional[int]:
        return self._value_of.get(metric)


def build_all_metrics(metrics_json: list[dict], optimizers: list[tuple[str, int]]) -> MetricIndex:
    defs: list[MetricDef] = []
    for group in metrics_json:
        gname = group.get("group")
        for metric in group.get("metrics", []):
            name = metric["name"]
            validate_identifier(name, gname)
            mtype = metric.get("type", "double")
            is_default = metric.get("is_default", False)
            query_root = "query_root" in metric
            collection_method = metric.get("collection_method")
            child = metric.get("child")
            description = metric.get("description", "")
            enum_value = metric.get("enum_value")
            defs.append(
                MetricDef(
                    name=name,
                    group=gname,
                    type=mtype if name != "OPERATOR_TYPE" else "uint8",
                    is_default=is_default,
                    query_root=query_root,
                    collection_method=collection_method,
                    child=child,
                    description=description,
                    enum_value=enum_value,
                )
            )
    return MetricIndex(defs, optimizers)



import os

common_path = os.path.join('src', 'include', 'duckdb', 'common', 'common.hpp')
with open(common_path, 'r') as f:
    text = f.read()


text = text.replace(
    '#pragma once',
    '''#pragma once

#ifdef _WIN32
#ifdef DUCKDB_MAIN_LIBRARY
#include "duckdb/common/windows.hpp"
#endif
#endif
''',
)

with open(common_path, 'w+') as f:
    f.write(text)



import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.csv as csv
from pathlib import Path


def generate_parquet(data_dir: Path):
    generate_silly_names(data_dir / 'silly-names.parquet')
    generate_byte_stream_split(data_dir / 'byte_stream_split.parquet')


def generate_silly_names(path: Path):
    df = pd.DataFrame({'önë': [1, 2, 3],
                       '': ['foo', 'bar', 'baz'],
                       '🦆': [True, False, True]})
    table = pa.Table.from_pandas(df)
    pq.write_table(table, path)


def generate_byte_stream_split(path: Path):
    num_rows = 100
    rng = np.random.default_rng(0)

    floats = pa.array(rng.uniform(-100.0, 100.0, num_rows), type=pa.float32())
    doubles = pa.array(rng.uniform(-100.0, 100.0, num_rows), type=pa.float64())

    null_mask = np.ones(num_rows, dtype=np.bool_)
    null_mask[num_rows // 10:] = False
    rng.shuffle(null_mask)
    nullable_floats = pa.array(
            rng.uniform(-100.0, 100.0, num_rows), type=pa.float32(), mask=null_mask)

    table = pa.Table.from_arrays(
            [floats, doubles, nullable_floats],
            ["floats", "doubles", "nullable_floats"])

    with pq.ParquetWriter(
            path,
            table.schema,
            use_dictionary=False,
            use_byte_stream_split=True) as writer:
        writer.write_table(table)

    csv_path = path.with_suffix('.csv')
    options = csv.WriteOptions(include_header=True, delimiter='|')
    csv.write_csv(table, csv_path, options)
    fix_csv_nulls(csv_path)


def fix_csv_nulls(path: Path):
    """ Replace empty values with 'NULL' """
    with open(path, 'r') as f:
        lines = f.readlines()

    with open(path, 'w') as f:
        for line in lines:
            split_line = ["NULL" if val == "" else val for val in line.strip().split("|")]
            f.write("|".join(split_line) + "\n")


if __name__ == '__main__':
    data_dir = Path(__file__).parent
    generate_parquet(data_dir)



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os

def test_version_dev(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_dev.db")
    )
    result = test.run()
    result.check_stderr("older development version")

def test_version_0_3_1(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_031.db")
    )
    result = test.run()
    result.check_stderr("v0.3.1")

def test_version_0_3_2(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_032.db")
    )
    result = test.run()
    result.check_stderr("v0.3.2")

def test_version_0_4(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_04.db")
    )
    result = test.run()
    result.check_stderr("v0.4.0")

def test_version_0_5_1(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_051.db")
    )
    result = test.run()
    result.check_stderr("v0.5.1")

def test_version_0_6_0(shell):
    test = (
        ShellTest(shell)
        .statement(".open test/storage/bc/db_060.db")
    )
    result = test.run()
    result.check_stderr("v0.6.0")

# fmt: on



from .parse_and_sort_settings_in_json import add_all_settings_to_global_list as parse_and_sort_json_file
from .update_settings_header_file import generate as update_header_file
from .update_settings_scopes import generate as update_scopes
from .update_settings_src_code import generate as update_src_code



import subprocess
import sys
import os

if len(sys.argv) < 3:
    print("Usage: [prefix] [filename1] [filename2] ... ")
    exit(1)


def git_rev_hash():
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode("utf-8").strip()


prefix = sys.argv[1].strip()

# Hannes controls this web server
# Files are served at https://download.duckdb.org/...

secret_key = os.getenv('DAV_PASSWORD')
if secret_key is None:
    print("Can't find DAV_PASSWORD in env ")
    exit(2)


files = sys.argv[2:]
basenames = [os.path.basename(x) for x in files]

basenames_set = set(basenames)
if len(basenames) != len(basenames_set):
    print("Name conflict")
    exit(3)


git_hash = git_rev_hash()

folder = 'rev/%s/%s' % (git_hash, prefix)


def curlcmd(cmd, path):
    p = subprocess.Popen(
        ['curl', '--retry', '10']
        + cmd
        + ['http://duckdb:%s@dav10635776.mywebdav.de/duckdb-download/%s' % (secret_key, path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    output, err = p.communicate()
    rc = p.returncode

    if p.returncode != 0:
        print(err)
        exit(4)


# create dirs, no recursive create supported by webdav
f_p = ''
for p in folder.split('/'):
    f_p += p + '/'
    curlcmd(['-X', 'MKCOL'], f_p)

for f in files:
    base = os.path.basename(f)
    key = '%s/%s' % (folder, base)
    print("%s\t->\thttps://download.duckdb.org/%s " % (f, key))
    curlcmd(['-T', f], key)



import os
import csv
import re
import argparse
import glob

os.chdir(os.path.dirname(__file__))

# Dont generate serialization for these enums
blacklist = [
    "RegexOptions",
    "Flags",
    "ContainerType",
    "Type",
    "DictionaryAppendState",
    "DictFSSTMode",
    "ComplexJSONType",
    "UnavailableReason",
    "Slot",
]

enum_util_header_file = os.path.join("..", "src", "include", "duckdb", "common", "enum_util.hpp")
enum_util_source_file = os.path.join("..", "src", "common", "enum_util.cpp")

# Overrides conversions for the following enums:
overrides = {
    "LogicalTypeId": {
        "SQLNULL": "NULL",
        "TIMESTAMP_TZ": "TIMESTAMP WITH TIME ZONE",
        "TIME_TZ": "TIME WITH TIME ZONE",
        "TIMESTAMP_SEC": "TIMESTAMP_S",
    },
    "JoinType": {"OUTER": "FULL"},
    "OrderType": {
        "ORDER_DEFAULT": ["ORDER_DEFAULT", "DEFAULT"],
        "DESCENDING": ["DESCENDING", "DESC"],
        "ASCENDING": ["ASCENDING", "ASC"],
    },
    "AllowParserOverride": {
        "DEFAULT_OVERRIDE": "DEFAULT",
        "FALLBACK_OVERRIDE": "FALLBACK",
        "STRICT_OVERRIDE": "STRICT",
    },
    "OrderByNullType": {
        "ORDER_DEFAULT": ["ORDER_DEFAULT", "DEFAULT"],
        "NULLS_FIRST": ["NULLS FIRST", "NULLS_FIRST"],
        "NULLS_LAST": ["NULLS LAST", "NULLS_LAST"],
    },
    "CheckpointAbort": {
        "NO_ABORT": "NONE",
        "DEBUG_ABORT_BEFORE_TRUNCATE": "BEFORE_TRUNCATE",
        "DEBUG_ABORT_BEFORE_HEADER": "BEFORE_HEADER",
        "DEBUG_ABORT_AFTER_FREE_LIST_WRITE": "AFTER_FREE_LIST_WRITE",
        "DEBUG_ABORT_BEFORE_WAL_FINISH": "BEFORE_WAL_FINISH",
        "DEBUG_ABORT_BEFORE_MOVING_RECOVERY": "BEFORE_MOVING_RECOVERY",
        "DEBUG_ABORT_BEFORE_DELETING_CHECKPOINT_WAL": "BEFORE_DELETING_CHECKPOINT_WAL",
    },
    "SampleMethod": {"SYSTEM_SAMPLE": "System", "BERNOULLI_SAMPLE": "Bernoulli", "RESERVOIR_SAMPLE": "Reservoir"},
    "TableReferenceType": {"EMPTY_FROM": "EMPTY"},
    "LogLevel": {
        "LOG_TRACE": "TRACE",
        "LOG_DEBUG": "DEBUG",
        "LOG_INFO": "INFO",
        "LOG_WARNING": "WARNING",
        "LOG_ERROR": "ERROR",
        "LOG_FATAL": "FATAL",
    },
    "RequestType": {
        "GET_REQUEST": "GET",
        "PUT_REQUEST": "PUT",
        "HEAD_REQUEST": "HEAD",
        "DELETE_REQUEST": "DELETE",
        "POST_REQUEST": "POST",
    },
    "CompressionType": {
        "COMPRESSION_AUTO": "AUTO",
        "COMPRESSION_UNCOMPRESSED": "UNCOMPRESSED",
        "COMPRESSION_CONSTANT": "CONSTANT",
        "COMPRESSION_RLE": "RLE",
        "COMPRESSION_DICTIONARY": "DICTIONARY",
        "COMPRESSION_PFOR_DELTA": "PFOR",
        "COMPRESSION_BITPACKING": "BITPACKING",
        "COMPRESSION_FSST": "FSST",
        "COMPRESSION_CHIMP": "CHIMP",
        "COMPRESSION_PATAS": "PATAS",
        "COMPRESSION_ALP": "ALP",
        "COMPRESSION_ALPRD": "ALPRD",
        "COMPRESSION_ZSTD": "ZSTD",
        "COMPRESSION_ROARING": "ROARING",
        "COMPRESSION_EMPTY": "EMPTY",
        "COMPRESSION_DICT_FSST": "DICT_FSST",
    },
    "ArrowFormatVersion": {"V1_0": "1.0", "V1_1": "1.1", "V1_2": "1.2", "V1_3": "1.3", "V1_4": "1.4", "V1_5": "1.5"},
}

# get all the headers
hpp_files = []
for root, dirs, files in os.walk(os.path.join("..", "src")):
    for file in files:
        # Dont include the generated header itself recursively
        if file == "enum_util.hpp":
            continue
        if 'amalgamation' in root:
            continue

        if file.endswith(".hpp"):
            hpp_files.append(os.path.join(root, file))


def remove_prefix(str, prefix):
    if str.startswith(prefix):
        return str[len(prefix) :]
    return str


# get all the enum classes
enums = []
enum_paths = []
enum_path_set = set()

for hpp_file in hpp_files:
    with open(hpp_file, "r") as f:
        text = f.read()
        for res in re.finditer(r"enum class (\w*)\s*:\s*(\w*)\s*{((?:\s*[^}])*)}", text, re.MULTILINE):
            file_path = remove_prefix(os.path.relpath(hpp_file, os.path.join("..", "src")), "include/")
            enum_name = res.group(1)

            if enum_name in blacklist:
                print(f"Skipping {enum_name} because it is blacklisted")
                continue

            enum_type = res.group(2)

            enum_members = []
            # Capture All members: \w+(\s*\=\s*-?\w*)?
            # group one is the member name
            # group two is the member value
            # First clean group from comments
            s = res.group(3)
            s = re.sub(r"\/\/.*", "", s)
            s = re.sub(r"\/\*.*\*\/", "", s)

            enum_values = {}
            for member in re.finditer(r"(\w+)(\s*\=\s*-?\w*)?", s):
                key = member.group(1)
                strings = [key]
                if enum_name in overrides and key in overrides[enum_name]:
                    override = overrides[enum_name][key]
                    if isinstance(override, list):
                        print(f"Overriding {enum_name}::{key} to one of {override}")
                        strings = override
                    else:
                        print(f"Overriding {enum_name}::{key} to {override}")
                        strings = [override]

                if member.group(2):
                    # If the member has a value, make sure it isnt already covered by another member
                    # If it is, we cant do anything else than ignore it
                    value = remove_prefix(member.group(2).strip(), "=").strip()
                    if value not in enum_values and value not in dict(enum_members):
                        enum_members.append((key, strings))
                    else:
                        print(f"Skipping {enum_name}::{key} because it has a duplicate value {value}")
                else:
                    enum_members.append((key, strings))

            if not file_path in enum_path_set:
                enum_path_set.add(file_path)
                enum_paths.append(file_path)

            enums.append((enum_name, enum_type, enum_members))

enum_paths.sort()
enums.sort(key=lambda x: x[0])

header = """//-------------------------------------------------------------------------
// This file is automatically generated by scripts/generate_enum_util.py
// Do not edit this file manually, your changes will be overwritten
// If you want to exclude an enum from serialization, add it to the blacklist in the script
//
// Note: The generated code will only work properly if the enum is a top level item in the duckdb namespace
// If the enum is nested in a class, or in another namespace, the generated code will not compile.
// You should move the enum to the duckdb namespace, manually write a specialization or add it to the blacklist
//-------------------------------------------------------------------------\n\n
"""

# Write the enum util header
with open(enum_util_header_file, "w") as f:
    f.write(header)

    f.write('#pragma once\n\n')
    f.write('#include <stdint.h>\n')
    f.write('#include "duckdb/common/string.hpp"\n\n')

    f.write("namespace duckdb {\n\n")

    f.write(
        """struct EnumUtil {
    // String -> Enum
    template <class T>
    static T FromString(const char *value) = delete;

    template <class T>
    static T FromString(const string &value) { return FromString<T>(value.c_str()); }

    // Enum -> String
    template <class T>
    static const char *ToChars(T value) = delete;

    template <class T>
    static string ToString(T value) { return string(ToChars<T>(value)); }
};\n\n"""
    )

    # Forward declare all enums
    for enum_name, enum_type, _ in enums:
        f.write(f"enum class {enum_name} : {enum_type};\n\n")
    f.write("\n")

    # Forward declare all enum serialization functions
    for enum_name, enum_type, _ in enums:
        f.write(f"template<>\nconst char* EnumUtil::ToChars<{enum_name}>({enum_name} value);\n\n")
    f.write("\n")

    # Forward declare all enum dserialization functions
    for enum_name, enum_type, _ in enums:
        f.write(f"template<>\n{enum_name} EnumUtil::FromString<{enum_name}>(const char *value);\n\n")
    f.write("\n")

    f.write("}\n")


with open(enum_util_source_file, "w") as f:
    f.write(header)

    f.write('#include "duckdb/common/enum_util.hpp"\n')

    # Write the includes
    for enum_path in enum_paths:
        f.write(f'#include "{enum_path}"\n')
    f.write("\n")

    f.write("namespace duckdb {\n\n")

    for enum_name, enum_type, enum_members in enums:
        enum_string_array = "Get" + enum_name + "Values()"
        # Write the enum from string
        f.write(f"const StringUtil::EnumStringLiteral *{enum_string_array} {{\n")
        f.write(f"\tstatic constexpr StringUtil::EnumStringLiteral values[] {{\n")
        member_count = 0
        for key, strings in enum_members:
            for str_val in strings:
                if member_count != 0:
                    f.write(",\n")
                f.write(f"\t\t{{ static_cast<uint32_t>({enum_name}::{key}), \"{str_val}\" }}")
                member_count += 1
        f.write("\n\t};")
        f.write("\n\treturn values;")
        f.write("\n}\n\n")
        f.write(f"template<>\nconst char* EnumUtil::ToChars<{enum_name}>({enum_name} value) {{\n")
        f.write(
            f"\treturn StringUtil::EnumToString({enum_string_array}, {member_count}, \"{enum_name}\", static_cast<uint32_t>(value));\n"
        )
        f.write("}\n\n")

        # Write the string to enum
        f.write(f"template<>\n{enum_name} EnumUtil::FromString<{enum_name}>(const char *value) {{\n")
        f.write(
            f"\treturn static_cast<{enum_name}>(StringUtil::StringToEnum({enum_string_array}, {member_count}, \"{enum_name}\", value));"
        )
        f.write("\n}\n\n")

    f.write("}\n\n")



# fmt: off

import pytest
import subprocess
import sys
from typing import List
from conftest import ShellTest
import os
from pathlib import Path

def test_dump_create(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE TABLE a (i INTEGER);")
        .statement(".changes off")
        .statement("INSERT INTO a VALUES (42);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE TABLE a(i INTEGER)')
    result.check_stdout('COMMIT')

@pytest.mark.parametrize("pattern", [
    "a",
    "a%"
])
def test_dump_specific(shell, pattern):
    test = (
        ShellTest(shell)
        .statement("CREATE TABLE a (i INTEGER);")
        .statement(".changes off")
        .statement("INSERT INTO a VALUES (42);")
        .statement(f".dump {pattern}")
    )
    result = test.run()
    result.check_stdout('CREATE TABLE a(i INTEGER)')

# Original comment: more types, tables and views
def test_dump_mixed(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE TABLE a (d DATE, k FLOAT, t TIMESTAMP);")
        .statement("CREATE TABLE b (c INTEGER);")
        .statement(".changes off")
        .statement("INSERT INTO a VALUES (DATE '1992-01-01', 0.3, NOW());")
        .statement("INSERT INTO b SELECT * FROM range(0,10);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE TABLE a(d DATE, k FLOAT, t TIMESTAMP);')

def test_dump_blobs(shell):
    test = (
        ShellTest(shell)
        .statement("create table test(t VARCHAR, b BLOB);")
        .statement(".changes off")
        .statement("insert into test values('literal blob', '\\x07\\x08\\x09');")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout("'\\x07\\x08\\x09'")

def test_dump_newline(shell):
    test = (
        ShellTest(shell)
        .statement("create table newline_data as select concat(chr(10), '\n') s;")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout("concat")
    result.check_stdout("chr(10)")

def test_dump_indexes(shell):
    test = (
        ShellTest(shell)
        .statement("create table integer(i int);")
        .statement("create index i_index on integer(i);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout("CREATE INDEX i_index")

def test_dump_views(shell):
    test = (
        ShellTest(shell)
        .statement("create table integer(i int);")
        .statement("create view v1 as select * from integer;")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout("CREATE VIEW v1")

def test_dump_schema_qualified(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE SCHEMA other;")
        .statement("CREATE TABLE other.t_in_other(a INT);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS other;')
    result.check_stdout('CREATE TABLE other.t_in_other(a INTEGER);')
    result.check_stdout('COMMIT')

def test_dump_schema_with_data(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE SCHEMA test_schema;")
        .statement("CREATE TABLE test_schema.tbl(x INT, y VARCHAR);")
        .statement(".changes off")
        .statement("INSERT INTO test_schema.tbl VALUES (1, 'hello'), (2, 'world');")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS test_schema;')
    result.check_stdout('CREATE TABLE test_schema.tbl(x INTEGER, y VARCHAR);')
    result.check_stdout("INSERT INTO test_schema.tbl VALUES(1,'hello');")
    result.check_stdout('COMMIT')

def test_dump_multiple_schemas(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE SCHEMA s1;")
        .statement("CREATE SCHEMA s2;")
        .statement("CREATE TABLE s1.t1(a INT);")
        .statement("CREATE TABLE s2.t2(b INT);")
        .statement(".changes off")
        .statement("INSERT INTO s1.t1 VALUES (10);")
        .statement("INSERT INTO s2.t2 VALUES (20);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS s1;')
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS s2;')
    result.check_stdout('INSERT INTO s1.t1 VALUES(10);')
    result.check_stdout('INSERT INTO s2.t2 VALUES(20);')

def test_dump_quoted_schema(shell):
    test = (
        ShellTest(shell)
        .statement('CREATE SCHEMA "my-schema";')
        .statement('CREATE TABLE "my-schema"."my-table"(a INT);')
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS "my-schema";')
    result.check_stdout('CREATE TABLE "my-schema"."my-table"(a INTEGER);')

def test_dump_if_not_exists(shell):
    test = (
        ShellTest(shell)
        .statement("CREATE SCHEMA other;")
        .statement("CREATE TABLE IF NOT EXISTS other.tbl(x INT);")
        .statement(".changes off")
        .statement("INSERT INTO other.tbl VALUES (42);")
        .statement(".dump")
    )
    result = test.run()
    result.check_stdout('CREATE SCHEMA IF NOT EXISTS other;')
    result.check_stdout('INSERT INTO other.tbl VALUES(42);')
    result.check_stdout('COMMIT')



import os
import argparse
import subprocess
import shutil

parser = argparse.ArgumentParser(
    description='''Runs storage tests both with explicit one-initialization and with explicit zero-initialization, and verifies that the final storage files are the same.
The purpose of this is to verify all memory is correctly initialized before writing to disk - which prevents leaking of in-memory data in storage files by writing uninitialized memory to disk.'''
)
parser.add_argument('--unittest', default='build/debug/test/unittest', help='path to unittest', dest='unittest')
parser.add_argument(
    '--zero_init_dir',
    default='test_zero_init_db',
    help='directory to write zero-initialized databases to',
    dest='zero_init_dir',
)
parser.add_argument(
    '--standard_dir', default='test_standard_db', help='directory to write regular databases to', dest='standard_dir'
)

args = parser.parse_args()

test_list = [
    'test/sql/index/art/storage/test_art_checkpoint.test',
    'test/sql/storage/compression/simple_compression.test',
    'test/sql/storage/delete/test_store_deletes.test',
    'test/sql/storage/mix/test_update_delete_string.test',
    'test/sql/storage/nested/struct_of_lists_unaligned.test',
    'test/sql/storage/test_store_integers.test',
    'test/sql/storage/test_store_nulls_strings.test',
    'test/sql/storage/update/test_store_null_updates.test',
]


def run_test(args):
    res = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = res.stdout.decode('utf8').strip()
    stderr = res.stderr.decode('utf8').strip()
    if res.returncode != 0:
        print("Failed to run test!")
        print("----------COMMAND-----------")
        print(' '.join(args))
        print("----------STDOUT-----------")
        print(stdout)
        print("----------STDERR-----------")
        print(stderr)
        print("---------------------")
        exit(1)


header_size = 4096 * 3
block_size = 262144
checksum_size = 8


def handle_error(i, standard_db, zero_init_db, standard_data, zero_data):
    print("------------------------------------------------------------------")
    print(f"FAIL - Mismatch between one-initialized and zero-initialized databases at byte position {i}")
    print("------------------------------------------------------------------")
    print(f"One-initialized database {standard_db} - byte value {standard_data}")
    print(f"Zero-initialized database {zero_init_db} - byte value {zero_data}")
    if i < header_size:
        print("This byte is in the initial headers of the file")
    else:
        byte_pos = (i - header_size) % block_size
        if byte_pos >= checksum_size:
            print(
                f"This byte is in block id {(i - header_size) // block_size} at byte position {byte_pos - checksum_size} (position {byte_pos} including the block checksum)"
            )
        else:
            print(f"This byte is in block id {(i - header_size) // block_size} at byte position {byte_pos}")
            print("This is in the checksum part of the block")
    print("------------------------------------------------------------------")
    print(
        "This error likely means that memory was not correctly zero-initialized in a block before being written out to disk."
    )


def compare_database(standard_db, zero_init_db):
    with open(standard_db, 'rb') as f:
        standard_data = f.read()
    with open(zero_init_db, 'rb') as f:
        zero_data = f.read()
    if len(standard_data) != len(zero_data):
        print(
            f"FAIL - Length mismatch between database {standard_db} ({str(len(standard_data))}) and {zero_init_db} ({str(len(zero_data))})"
        )
        return False
    found_error = None
    for i in range(len(standard_data)):
        if standard_data[i] != zero_data[i]:
            if i > header_size:
                byte_pos = (i - header_size) % block_size
                if byte_pos <= 8:
                    # different checksum, skip because it does not tell us anything!
                    if found_error is None:
                        found_error = i
                    continue
            handle_error(i, standard_db, zero_init_db, standard_data[i], zero_data[i])
            return False
    if found_error is not None:
        i = found_error
        handle_error(i, standard_db, zero_init_db, standard_data[i], zero_data[i])
        return False
    print("Success!")
    return True


def compare_files(standard_dir, zero_init_dir):
    standard_list = os.listdir(standard_dir)
    zero_init_list = os.listdir(zero_init_dir)
    standard_list.sort()
    zero_init_list.sort()
    if standard_list != zero_init_list:
        print(
            f"FAIL - Directories contain mismatching files (standard - {str(standard_list)}, zero init - {str(zero_init_list)})"
        )
        return False
    if len(standard_list) == 0:
        print("FAIL - Directory is empty!")
        return False
    success = True
    for entry in standard_list:
        if not compare_database(os.path.join(standard_dir, entry), os.path.join(zero_init_dir, entry)):
            success = False
    return success


def clear_directories(directories):
    for dir in directories:
        try:
            shutil.rmtree(dir)
        except FileNotFoundError as e:
            pass


test_dirs = [args.standard_dir, args.zero_init_dir]

success = True
for test in test_list:
    print(f"Running test {test}")
    clear_directories(test_dirs)
    standard_args = [args.unittest, '--test-temp-dir', args.standard_dir, '--one-initialize', '--single-threaded', test]
    zero_init_args = [
        args.unittest,
        '--test-temp-dir',
        args.zero_init_dir,
        '--zero-initialize',
        '--single-threaded',
        test,
    ]
    print(f"Running test in one-initialize mode")
    run_test(standard_args)
    print(f"Running test in zero-initialize mode")
    run_test(zero_init_args)
    if not compare_files(args.standard_dir, args.zero_init_dir):
        success = False

clear_directories(test_dirs)

if not success:
    exit(1)



from bs4 import BeautifulSoup
import requests
import re


pages = [
    'avg',
    'count',
    'max',
    'min',
    'stdev',
    'sum',
    'var',
    'cume-dist',
    'first-value',
    'last-value',
    'lag',
    'lead',
    'percent-rank',
    'dense-rank',
    'ntile',
    'rank',
    'row-number',
]

# crash!
# url = 'https://docs.microsoft.com/en-us/sql/t-sql/functions/first-value-transact-sql?view=sql-server-2017'
url = 'https://docs.microsoft.com/en-us/sql/t-sql/functions/%s-transact-sql?view=sql-server-2017'


def transform_result_set(tblstr):
    # find the row with the --- --- etc that indicates col count
    cols = len(re.findall(r"-{3,}( |\n)", tblstr))
    if cols < 1:
        return ""
    print("REQUIRE(result->ColumnCount() == %d);" % cols)
    lineiterator = iter(tblstr.splitlines())
    in_data = False
    result = []
    for c in range(0, cols):
        result.append([])

    for line in lineiterator:
        if '---' in line:
            in_data = True
            continue
        if not in_data:
            continue
        if 'row(s) affected' in line:
            continue
        if re.match(r"^\s*$", line):
            continue

        #  now we have a real data line, split by space and trim
        fields = re.split(r"\s{2,}", line)
        if len(fields) < cols:
            raise ValueError('Not enough fields')

        # print("// ", end='')
        for c in range(0, cols):
            f = fields[c].strip()
            if f == '':
                raise ValueError('Empty field')
            if re.match(r"^\d+[\d,]*\.\d+$", f):
                f = f.replace(',', '')
            # 	print(f + '\t', end='')
            needs_quotes = False
            try:
                float(f)
            except ValueError:
                needs_quotes = True

            if f == "NULL":
                f = 'Value()'
                needs_quotes = False

            if needs_quotes:
                f = '"%s"' % f
            result[c].append(f)
        # print()

    for c in range(0, cols):
        print('REQUIRE(CHECK_COLUMN(result, %d, {%s}));' % (c, ','.join(result[c])))


for p in pages:
    r = requests.get(url % p)
    print('\n\n// FROM %s\n' % url % p)
    soup = BeautifulSoup(r.content, 'html.parser')
    look_for_answer = False

    for code in soup.find_all('code'):
        classes = code.get('class')
        text = code.get_text()
        if text.count('\n') < 2:
            continue
        if 'SELECT ' in text and 'FROM ' in text:
            if 'dbo.' in text or 'sys.' in text:
                continue
            query = text.strip()
            query = re.sub(r"(^|\n)(GO|USE|DECLARE).*", "", query)
            query = query.replace('\n', ' ')
            query = re.sub(r"\s+", " ", query)

            print('\n\nresult = con.Query("%s");\nREQUIRE(!result->HasError());' % query.replace('"', '\\"'))

            look_for_answer = True
        elif look_for_answer:
            # print('-- ' + text.replace('\n', '\n-- ') + '\n')
            transform_result_set(text)
            look_for_answer = False



import os
import subprocess

duckdb_program = '/Users/myth/Programs/duckdb-bugfix/build/release/duckdb'

struct_def = '''struct $STRUCT_NAME {
	static constexpr char *Name = "$NAME";
	static const char *Columns[];
	static constexpr idx_t ColumnCount = $COLUMN_COUNT;
	static const LogicalType Types[];
	static constexpr idx_t PrimaryKeyCount = $PK_COLUMN_COUNT;
	static const char *PrimaryKeyColumns[];
};
'''

initcode = '''
call dsdgen(sf=0);
.mode csv
.header 0
'''

column_count_query = '''
select count(*) from pragma_table_info('$NAME');
'''

pk_column_count_query = '''
select count(*) from pragma_table_info('$NAME') where pk=true;
'''

gen_names = '''
select concat('const char *', '$STRUCT_NAME', '::Columns[] = {', STRING_AGG('"' || name || '"', ', ') || '};') from pragma_table_info('$NAME');
'''

gen_types = '''
select concat('const LogicalType ', '$STRUCT_NAME', '::Types[] = {', STRING_AGG('LogicalType::' || type, ', ') || '};') from pragma_table_info('$NAME');
'''

pk_columns = '''
select concat('const char *', '$STRUCT_NAME', '::PrimaryKeyColumns[] = {', STRING_AGG('"' || name || '"', ', ') || '};') from pragma_table_info('$NAME') where pk=true;
'''


def run_query(sql):
    input_sql = initcode + '\n' + sql
    res = subprocess.run(duckdb_program, input=input_sql.encode('utf8'), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    stdout = res.stdout.decode('utf8').strip()
    stderr = res.stderr.decode('utf8').strip()
    if res.returncode != 0:
        print("FAILED TO RUN QUERY")
        print(stderr)
        exit(1)
    return stdout


def prepare_query(sql, table_name, struct_name):
    return sql.replace('$NAME', table_name).replace('$STRUCT_NAME', struct_name)


header = '''
#pragma once

#include "duckdb.hpp"

#ifndef DUCKDB_AMALGAMATION
#include "duckdb/common/exception.hpp"
#include "duckdb/common/types/date.hpp"
#include "duckdb/parser/column_definition.hpp"
#include "duckdb/storage/data_table.hpp"
#include "duckdb/catalog/catalog_entry/table_catalog_entry.hpp"
#include "duckdb/planner/parsed_data/bound_create_table_info.hpp"
#include "duckdb/parser/parsed_data/create_table_info.hpp"
#include "duckdb/parser/constraints/not_null_constraint.hpp"
#include "duckdb/catalog/catalog.hpp"
#include "duckdb/planner/binder.hpp"
#endif

namespace tpcds {

using duckdb::LogicalType;
using duckdb::idx_t;
'''

footer = '''
}
'''

print(header)

table_list = run_query('show tables')
for table_name in table_list.split('\n'):
    table_name = table_name.strip()
    print(
        '''
//===--------------------------------------------------------------------===//
// $NAME
//===--------------------------------------------------------------------===//'''.replace(
            '$NAME', table_name
        )
    )
    struct_name = str(table_name.title().replace('_', '')) + 'Info'
    column_count = int(run_query(prepare_query(column_count_query, table_name, struct_name)).strip())
    pk_column_count = int(run_query(prepare_query(pk_column_count_query, table_name, struct_name)).strip())
    print(
        prepare_query(struct_def, table_name, struct_name)
        .replace('$COLUMN_COUNT', str(column_count))
        .replace('$PK_COLUMN_COUNT', str(pk_column_count))
    )

    print(run_query(prepare_query(gen_names, table_name, struct_name)).replace('""', '"').strip('"'))
    print("")
    print(run_query(prepare_query(gen_types, table_name, struct_name)).strip('"'))
    print("")
    print(run_query(prepare_query(pk_columns, table_name, struct_name)).replace('""', '"').strip('"'))

print(footer)
